/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP.BaseDistributions;

/* Original code from turnip.pepclust package.
Modified: 
To allow for empty angles, makes variables private
No longer a subclass of Vector
Disallows observations outside of -pi, pi
 */
public class Angles {

    public final boolean isEmpty;
    private double phi_;
    private double psi_;
    private double phiInDegrees_;
    private double psiInDegrees_;
    private double cosPhi_;
    private double cosPsi_;
    private double sinPhi_;
    private double sinPsi_;
    
    /**
     *  Defaul constructor, creates an empty angle.
     */
    public Angles() {
        isEmpty = true;
    }

    /**
     *  Standard constructor, takes phi, psi in radians
     * @param phi
     * @param psi
     */
    public Angles(double phi, double psi) {
        isEmpty = false;
        // Automatically wraps between -pi and pi
        phi_ = Math.atan2(Math.sin(phi), Math.cos(phi));
        psi_ = Math.atan2(Math.sin(psi), Math.cos(psi));
        phiInDegrees_ = phi_ * TO_DEGREES;
        psiInDegrees_ = psi_ * TO_DEGREES;
        cosPhi_=Math.cos(phi_);
        cosPsi_=Math.cos(psi_);
        sinPhi_=Math.sin(phi_);
        sinPsi_=Math.sin(psi_);

    }

    /**
     * Takes phi, psi in radians or degrees
     * @param phi
     * @param psi
     * @param inDegrees
     */
    public Angles(double phi, double psi, boolean inDegrees) {
        if (inDegrees) {
            phi *= TO_RADIANS;
            psi *= TO_RADIANS;
        }

        isEmpty = false;
        // Automatically wraps between -pi and pi
        phi_ = Math.atan2(Math.sin(phi), Math.cos(phi));
        psi_ = Math.atan2(Math.sin(psi), Math.cos(psi));
        phiInDegrees_ = phi_ * TO_DEGREES;
        psiInDegrees_ = psi_ * TO_DEGREES;
        cosPhi_=Math.cos(phi_);
        cosPsi_=Math.cos(psi_);
        sinPhi_=Math.sin(phi_);
        sinPsi_=Math.sin(psi_);
    }

// Accessor method
    public double phi() throws Exception {
        if (!isEmpty) {
            return phi_;
        } else {
            throw new Exception("Empty Vector.");
        }
    }

    public double phi(boolean inDegrees) throws Exception {
        if (!isEmpty) {
            if (inDegrees) {
                return phiInDegrees_;
            } else {
                return phi_;
            }
        } else {
            throw new Exception("Empty Vector.");
        }
    }

    public double psi() throws Exception {
        if (!isEmpty) {
            return psi_;
        } else {
            throw new Exception("Empty Vector.");
        }
    }

    public double psi(boolean inDegrees) throws Exception {
        if (!isEmpty) {
            if (inDegrees) {
                return psiInDegrees_;
            } else {
                return psi_;
            }
        } else {
            throw new Exception("Empty Vector.");
        }
    }

    
    public double cosPsi() throws Exception {
        if (!isEmpty) {
                return cosPsi_;
            
        } else {
            throw new Exception("Empty Vector.");
        }
    }

        public double cosPhi() throws Exception {
        if (!isEmpty) {
                return cosPhi_;
            
        } else {
            throw new Exception("Empty Vector.");
        }
    }
       
    public double sinPsi() throws Exception {
        if (!isEmpty) {
                return sinPsi_;
            
        } else {
            throw new Exception("Empty Vector.");
        }
    }
    
      
    public double sinPhi() throws Exception {
        if (!isEmpty) {
                return sinPhi_;
            
        } else {
            throw new Exception("Empty Vector.");
        }
    }
    
    public String toString() {
        return toString(false);
    }

    public String toString(boolean inDegrees) {
        if (this.isEmpty) {
            return ("Empty.");
        }
        if (inDegrees) {
            return String.format("%5.4f %5.4f", phiInDegrees_, psiInDegrees_);
        } else {
            return String.format("%5.4f %5.4f", phi_, psi_);
        }
    }
    
    /**
     *  Recenter angle pair. Returns new Angles object.
     *  Will never actually throw an exception.
     */
    public Angles subtract(Angles temp) throws Exception{
        if((!isEmpty)&&(!temp.isEmpty))
            return new Angles(phi_-temp.phi(), psi_-temp.psi());
        else
            return new Angles();
    }
    
    public static final double TO_DEGREES = 180.0 / Math.PI;
    public static final double TO_RADIANS = Math.PI / 180.0;
}
