/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP.BaseDistributions;


import HMMDP.*;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math.special.Gamma;
import org.apache.commons.math.util.MathUtils;


// ORIG COMMENTS FROM DBD
// WARNING:  My parametrization of lambda is the negative of that of Lennox, et. al. (2008) and Mardia (2007)!
// Modified to use KPL Angles class; otherwise unchanged

// Possible future changes: allow empty angles in logDistribution, return 0.0
// When all precision parameters are 0, behaves as uniform on [-pi,pi)x[-pi,pi)

public class BivariateVonMises implements Distribution<Angles> {

  // Parameters
  private Angles mean;
  private Scale scale;

  // Normalizing constant corresponding to scale parameters, and indicator of whether it's been calculated
  private boolean cached;
  private double cachedLogNormalizingConstant;

  // General math class constants
  private static final double TWO_PI = 2*Math.PI;
  private static final double FOUR_PI_SQ = TWO_PI * TWO_PI;
  private static final double MINUS_LOG_FOUR_PI_SQ = -Math.log( FOUR_PI_SQ );



  // Class constants for rapid computation of normalizing constant
  private static final double[] logGamma;
  private static final double[] functionOfBinomialCoefficient;
  private static final int INDEX_MAX = 1000;

  // Initializing class constants (see above)
  static {
    logGamma = new double[INDEX_MAX];
    functionOfBinomialCoefficient = new double[INDEX_MAX];
    for ( int i=0; i<INDEX_MAX; i++ ) {
      logGamma[i] = Gamma.logGamma(i);
      functionOfBinomialCoefficient[i] = MathUtils.binomialCoefficientDouble(2*i,i);
    }
  }

  public BivariateVonMises(){
  }
  
  public BivariateVonMises(Angles mean, Scale precision) throws Exception {
    this.mean = mean;
    if(this.mean.isEmpty) throw new Exception ("Empty Angles not allowed as parameters.");
    this.scale = precision;
    cached = false;
  }
  
  public BivariateVonMises(ParameterSet params) throws Exception{
      this.mean=params.center;
      if(this.mean.isEmpty) throw new Exception ("Empty Angles not allowed as parameters.");
      this.scale=params.scale;
      cached=false;
  }

  // With normalizing constant
  // Return 0 makes consistent with existing logLikelihood function.
  public double logDensity(Angles x) {
      if(x.isEmpty)
          return 0.0;
      else
          return logNormalizingConstant() + logDensityKernel(x);
  }

  // Without normalizing constant
  public double logDensityKernel(Angles x) {
        if(scale.phi==0.0&&scale.psi==0&&scale.lambda==0.0)
            return MINUS_LOG_FOUR_PI_SQ;
        try {
            double differencePhi = x.phi() - mean.phi();
            double differencePsi = x.psi() - mean.psi();
            return scale.phi * Math.cos(differencePhi) + scale.psi * Math.cos(differencePsi) - scale.lambda * Math.sin(differencePhi) * Math.sin(differencePsi);
        } catch (Exception ex) {
            Logger.getLogger(BivariateVonMises.class.getName()).log(Level.SEVERE, null, ex);
            return 0.0;
        }
  }

  public double logLikelihood(Angles[] data){
      double temp = 0.0;
      for(Angles x:data){
          if(!x.isEmpty)
              temp+=logDensity(x);        
      }
      return temp;
  }
  

  public double logLikelihoodKernel(Angles[] data){
      double temp = 0.0;
      for(Angles x:data){
          if(!x.isEmpty)
              temp+=logDensityKernel(x);        
      }
      return temp;
  }
  
    
  
  // Modified Bessel function of the first kind of order 'order' evaluated at 'x'
  private double besselIn(int order, double x) {
    final double a = Math.log(x/2.0);
    int m = 0;
    double y = 0.0;
    while ( true ) {
      if((m+order+1)>=INDEX_MAX)
          break;
      double newY = y + Math.exp( (2*m+order)*a - logGamma[m+1] - logGamma[m+order+1] );
      if ( newY == y) {
        break;
      }
      y = newY;
      m++;
    }
    return y;
  }

  public double logNormalizingConstant() {
    // If uniform
    if(scale.phi==0.0&&scale.psi==0.0&&scale.lambda==0.0)
        return 0.0;
    // If cached return value
    if ( cached ) return cachedLogNormalizingConstant;
    // Else calculate and cache
    final double a = scale.lambda*scale.lambda/(4.0*scale.phi*scale.psi);
    int m = 0;
    double y = 0.0;
    double pow = 1.0;
    while ( true ) {
      if(m>=INDEX_MAX)
          break;
      double newY = y + functionOfBinomialCoefficient[m] * pow * besselIn(m,scale.phi) * besselIn(m,scale.psi);
      //if ( newY == y ) {
      if ( newY / y < 1.0001 ) {
        break;
      }
      y = newY;
      pow *= a;
      m++;
    }
    cachedLogNormalizingConstant = MINUS_LOG_FOUR_PI_SQ - Math.log(y);
    cached = true;
    return cachedLogNormalizingConstant;
  }


  // Rejection sampler modified so that Angles class does not require a sampling method
  public Angles sample(Random random) {
    final double logMax = logDensityKernel(mode());
      // If uniform
    if(scale.phi==0.0&&scale.psi==0.0&&scale.lambda==0.0)
       return new Angles((random.nextDouble()-.5)*TWO_PI,(random.nextDouble()-.5)*2*TWO_PI);
    while ( true ) {
      Angles candidate = new Angles((random.nextDouble()-.5)*TWO_PI,(random.nextDouble()-.5)*2*TWO_PI);
      if ( Math.log(random.nextDouble()) < logDensityKernel(candidate) - logMax ) return candidate;
    }
  }

  // Check for unimodality
  public boolean unimodal() {
    return scale.phi * scale.psi > scale.lambda * scale.lambda;
  }

  // Finds coodinates of mode. Modified to use KPL Angles class.
  // If an empty angles exception is caught, returns empty Angles object. Should only happen
  // if mean vector is improperly defined.
  public Angles mode() {
    if ( unimodal() ) return mean;
    double lambdaAbs = Math.abs(scale.lambda);
    double lambdaSq = lambdaAbs*lambdaAbs;
    double a = Math.sqrt( ( lambdaSq + scale.phi*scale.phi ) / ( lambdaSq + scale.psi*scale.psi ) );
    double cosPsiP = Math.abs(scale.psi) / lambdaAbs * a;
    double cosPsiN = -cosPsiP;
    double cosPhiP = Math.abs(scale.phi) / lambdaAbs / a;
    double cosPhiN = -cosPhiP;
    double psiP = Math.acos(cosPsiP);
    double psiN = Math.acos(cosPsiN);
    double phiP = Math.acos(cosPhiP);
    double phiN = Math.acos(cosPhiN);
    Angles candidate1;
    Angles candidate2;
    if ( scale.lambda < 0.0 ) {
            try {
                candidate1 = new Angles(mean.phi() + phiP, mean.psi() + psiP);
                candidate2 = new Angles(mean.phi() + phiN, mean.psi() + psiN);
                if ( logDensityKernel(candidate1) > logDensityKernel(candidate2) ) return candidate1;
                else return candidate2;
            } catch (Exception ex) {
                Logger.getLogger(BivariateVonMises.class.getName()).log(Level.SEVERE, null, ex);
                return new Angles();
            }
    } else {
            try {
                candidate1 = new Angles(mean.phi() + phiN, mean.psi() + psiP);
                candidate2 = new Angles(mean.phi() + phiP, mean.psi() + psiN);
                if ( logDensityKernel(candidate1) > logDensityKernel(candidate2) ) return candidate1;
                else return candidate2;
            } catch (Exception ex) {
                Logger.getLogger(BivariateVonMises.class.getName()).log(Level.SEVERE, null, ex);
                return new Angles();
            }
      
    }
   
  }

  // Given a data set and precision matrix, returns a proposal distribution
  // for the mean, assuming each observation is i.i.d. bivariate von Mises, and
  // a uniform prior on the mean. NOT A TRUE FULL CONDITIONAL.
  public static BivariateVonMises fullConditionalUniform(Angles[] data, Scale precision) throws Exception{
      double lambda = precision.lambda;
      double k1=precision.phi;
      double k2=precision.psi;
      
      double x1=0.0;
      double x2=0.0;
      double y1=0.0;
      double y2=0.0;
      double xy1=0.0;
      double xy2=0.0;
      
      
      for(int i=0;i<data.length;i++){
          // Won't throw exception due to this check.
          if(!data[i].isEmpty){
              x1+=data[i].cosPhi();
              x2+=data[i].sinPhi();
              y1+=data[i].cosPsi();
              y2+=data[i].sinPsi();
              xy1+=data[i].cosPhi()*data[i].cosPsi();
              xy2+=data[i].sinPhi()*data[i].sinPsi();
          }
      }
      
      x1*=k1;
      x2*=k1;
      y1*=k2;
      y2*=k2;
      double k1New = Math.sqrt(x1*x1+x2*x2);
      double k2New = Math.sqrt(y1*y1+y2*y2);
      Angles newCenter = new Angles(Math.atan2(x2,x1),Math.atan2(y2,y1));
      double lambdaNew = (xy1+xy2)*lambda/(x1*y1+x2*y2)*k1New*k2New;
      Scale newPrecision = Scale.precision(k1New, k2New, lambdaNew);
      return new BivariateVonMises(newCenter, newPrecision);
  }

  
  
  // Returns a proposal distribution for the mean if this is prior.
  // Assumes one precision matrix per cluster. NOT A TRUE FULL CONDITIONAL.
  public BivariateVonMises fullConditional(Angles[] data, Scale precision) throws Exception{
      double lambda = precision.lambda;
      double k1=precision.phi;
      double k2=precision.psi;
      
      double x1=0.0;
      double x2=0.0;
      double y1=0.0;
      double y2=0.0;
      double xy1=0.0;
      double xy2=0.0;
      
      
      for(int i=0;i<data.length;i++){
          // Won't throw exception due to this check.
          if(!data[i].isEmpty){
              x1+=data[i].cosPhi();
              x2+=data[i].sinPhi();
              y1+=data[i].cosPsi();
              y2+=data[i].sinPsi();
              xy1+=data[i].cosPhi()*data[i].cosPsi();
              xy2+=data[i].sinPhi()*data[i].sinPsi();
          }
      }
      x1*=k1;
      x2*=k1;
      y1*=k2;
      y2*=k2;
      x1+=scale.phi*mean.cosPhi();
      x2+=scale.phi*mean.sinPhi();
      y1+=scale.psi*mean.cosPsi();
      y2+=scale.psi*mean.sinPsi();
      double k1New = Math.sqrt(x1*x1+x2*x2);
      double k2New = Math.sqrt(y1*y1+y2*y2);
      Angles newCenter = new Angles(Math.atan2(x2,x1),Math.atan2(y2,y1));
      double lambdaNew = ((xy1+xy2)*lambda+scale.lambda*(mean.cosPhi()*mean.cosPsi()+mean.sinPhi()*mean.sinPsi()))/(x1*y1+x2*y2)*k1New*k2New;
      Scale newPrecision = Scale.precision(k1New, k2New, lambdaNew);
      return new BivariateVonMises(newCenter, newPrecision);
  }
   
}
