/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP.BaseDistributions;



import HMMDP.*;
import java.util.Random;

/**
 * Finite mixture of von Mises distributions. 
 * @author lennox
 */
public class BivariateVonMisesMix extends BivariateVonMises{

    private final BivariateVonMises[] components;
    private final double[] probs;
    
    public BivariateVonMisesMix(BivariateVonMises[] components, double[] probs) throws Exception{
        if(components.length!=probs.length)
            throw new Exception("Distribution array and probability array have diverent dimensions" );
        this.components=components;
        this.probs=probs;
        
        // Ensure probs sums to 1
        double temp=0.0;
        for(int i=0;i<this.probs.length;i++){
            temp+=this.probs[i];
            if(this.probs[i]<0) throw new Exception("Components of probability array must be non-negative.");
        }
        for(int i=0;i<this.probs.length;i++){
            this.probs[i]/=temp;
        }
    }
    
    @Override
    public double logDensity(Angles x) {
        double temp=0.0;
        for(int i=0;i<probs.length;i++){
            temp+=Math.exp(components[i].logDensity(x)*probs[i]);
        }
        return(Math.log(temp));
    }

    // Same as logDensity
    @Override
    public double logDensityKernel(Angles x) {
         return logDensity(x);
    }

    // None. logDensity and logDensityKernel are the same here.
    @Override
    public double logNormalizingConstant() {
        return 0.0;
    }

    @Override
    public Angles sample(Random random) {
        return(components[randomInt(probs, random)].sample(random));
    }
      

    // Method takes an array of probabilities and returns an integer with probability corresponding
    // to the value at that position in the array
    private int randomInt(double[] probArray, Random random) {
        // Clone probability array. Allows us to ensure that entries sum to one. Find scaling factor, and check that all
        // values are > 0
        double[] tempProbs = (double[]) (probArray.clone());
        double total = 0.0;
        for (int i = 0; i < tempProbs.length; i++) {
            if (tempProbs[i] < 0) {
                throw new IllegalArgumentException("Elements of probability array must be non-negative.");
            }
            total += tempProbs[i];
        }
        // Rescale probabilities to sum to 1, and change from density cdf
        tempProbs[0] /= total;
        for (int i = 1; i < tempProbs.length; i++) {
            tempProbs[i] = tempProbs[i] / total + tempProbs[i - 1];
        }

        double value = random.nextDouble();
        int index = 0;
        while (tempProbs[index] <= value) {
            index++;
        }
        return (index);
    }


  /**
   *  Returns the correct full conditional distribution for the mean given that
   *  all data comes from biv von Mises distributions with a known scale matrix,
   *  and the calling distribution as the prior on the mean
   */
    @Override
  public BivariateVonMisesMix fullConditional(Angles[] data, Scale precision) throws Exception{
      BivariateVonMises[] compCond = new BivariateVonMises[components.length];
      double[] newProbs = new double[probs.length];
      double temp = 0.0;
      for(int i=0;i<probs.length;i++){
          compCond[i] = components[i].fullConditional(data, precision);
          newProbs[i] = probs[i]*Math.exp(components[i].logNormalizingConstant()-compCond[i].logNormalizingConstant());
          temp+=newProbs[i];
      }
      // Ensuring probs sum to 1
      for(int i=0;i<probs.length;i++){
          newProbs[i]/=temp;
      }
      return new BivariateVonMisesMix(compCond,newProbs);      
  }
}
