/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP.BaseDistributions;

// Originally in DBD turnip.pepclust.
// Modified to use KPL Vector class
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math.special.Gamma;

// Note:  This is Bernardo & Smith's parametrization with E(W) = a * inv(B).
// The alternative parametrization has E(W) = v*S, where v = 2a and S = inv(2*B).
public class BivariateWishart implements Distribution<Scale> {

    public final double a;
    public final Scale B;

    public BivariateWishart(double a, Scale B) {
        this.a = a;
        this.B = B;
        if (a <= 0.5) {
            throw new IllegalArgumentException("a must be greater than 0.5.");
        }
    }

    public static BivariateWishart alternativeParametrization(double v, Scale S) {
        return new BivariateWishart(v / 2.0, Scale.precision(S.var1 / 2.0, S.var2 / 2.0, S.covar / 2.0));
    }

    @Override
    public Object clone() {
        Object result = null;
        try {
            result = super.clone();
        } catch (CloneNotSupportedException e) {
        }
        return result;
    }

    // With normalizing constant
    public double logDensity(Scale x) {
        return logNormalizingConstant() + logDensityKernel(x);
    }

    // Without normalizing constant
    public double logDensityKernel(Scale x) {
        return (a - 1.5) * Math.log(x.phi * x.psi - x.lambda * x.lambda) - (x.phi * B.phi + x.psi * B.psi + 2.0 * x.lambda * B.lambda);
    }

    public double logNormalizingConstant() {
        return MINUS_HALF_LOG_PI + a * Math.log(B.phi * B.psi - B.lambda * B.lambda) - Gamma.logGamma(a) - Gamma.logGamma(a - 0.5);
    }

    public Scale sample(Random random) {
        final BivariateNormal bn = new BivariateNormal(Scale.precision(2 * B.phi, 2 * B.psi, 2 * B.lambda));
        final int n = (int) Math.floor(2 * a + 1);
        double phi1 = 0;
        double psi1 = 0;
        double phi2 = 0;
        double psi2 = 0;
        double lambda2 = 0;
        for (int i = 0; i < n; i++) {
            try {
                Vector x = bn.sample(random);
                phi1 += x.phi();
                psi1 += x.psi();
                phi2 += x.phi() * x.phi();
                psi2 += x.psi() * x.psi();
                lambda2 += x.phi() * x.psi();
            } catch (Exception ex) {
                Logger.getLogger(BivariateWishart.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        phi2 -= phi1 * phi1 / n;
        psi2 -= psi1 * psi1 / n;
        lambda2 -= phi1 * psi1 / n;
        return Scale.precision(phi2, psi2, lambda2);
    }

    // Give an array of n iid biv Wishart dists
    public Scale[] sample(int n, Random random) {
        Scale[] temp = new Scale[n];
        for (int i = 0; i < n; i++) {
            temp[i] = this.sample(random);
        }
        return temp;
    }

    // Returns a proposal  distribution for the scale if this is prior
    // For use with angles class.
    // NOT A TRUE FULL CONDITIONAL.
    public BivariateWishart fullConditional(Angles[] data, Angles center) throws Exception {
        // Count non-empty observations
        int count = 0;
        double phi = 0.0;
        double psi = 0.0;
        double lambda = 0.0;
        for (Angles point : data) {
            if (!point.isEmpty) {
                count++;
                Angles temp = point.subtract(center);
                phi += temp.phi() * temp.phi();
                psi += temp.psi() * temp.psi();
                lambda += temp.phi() * temp.psi();
            }
        }
        int hi=0;
      //  for(int i=0;i<data.length;i++)
      //      if(!data[i].isEmpty)
      //          hi++;
      //  System.out.println(data.length+" "+count+" "+hi);
        Scale scale = Scale.precision(B.phi + phi, B.psi + psi, B.lambda + lambda);
        return new BivariateWishart(a + count / 2.0, scale);
    }
    private final double MINUS_HALF_LOG_PI = -0.5 * Math.log(Math.PI);
}
