/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP.BaseDistributions;

// Orig from turnip.pepclust package

public class Scale {

  public final double phi;
  public final double psi;
  public final double lambda;

  public final double var1;
  public final double var2;
  public final double covar;

  public final double std1;
  public final double std2;
  public final double corr;

  private Scale(double phi, double psi, double lambda, double var1, double var2, double covar, double std1, double std2, double corr) {
    this.phi = phi;
    this.psi = psi;
    this.lambda = lambda;
    this.var1 = var1;
    this.var2 = var2;
    this.covar = covar;
    this.std1 = std1;
    this.std2 = std2;
    this.corr = corr;
  }

  public static Scale precision(double phi, double psi, double lambda) {
    if ( (phi <= 0.0) || (psi <= 0.0) || Double.isInfinite(lambda) || Double.isNaN(lambda) ) {
      throw new IllegalArgumentException();
    }
    double determinant = phi*psi - lambda*lambda;
    double var1 = psi/determinant;
    double var2 = phi/determinant;
    double covar = -lambda/determinant;
    double std1 = Math.sqrt(var1);
    double std2 = Math.sqrt(var2);
    double corr = covar / (std1*std2);
    return new Scale(phi,psi,lambda,var1,var2,covar,std1,std2,corr);
  }

  public static Scale standardDeviation(double std1, double std2, double corr) {
    if ( (std1 <= 0.0) || (std2 <= 0.0) || (corr <= -1.0) || (corr >= 1.0) ) {
      throw new IllegalArgumentException();
    }
    double var1 = std1*std1;
    double var2 = std2*std2;
    double covar = corr*std1*std2;
    double determinant = var1 * var2 - covar * covar;
    double phi = var2 / determinant;
    double psi = var1 / determinant;
    double lambda = -covar / determinant;
    return new Scale(phi,psi,lambda,var1,var2,covar,std1,std2,corr);
  }

  public static Scale variance(double var1, double var2, double covar) {
    if ( (var1 <= 0.0) || (var2 <= 0.0) || ( Math.abs(covar) >= Math.sqrt(var1*var2) ) ) {
      throw new IllegalArgumentException();
    }
    double std1 = Math.sqrt(var1);
    double std2 = Math.sqrt(var2);
    double corr = covar / ( std1 * std2 );
    double determinant = var1 * var2 - covar * covar;
    double phi = var2 / determinant;
    double psi = var1 / determinant;
    double lambda = -covar / determinant;
    return new Scale(phi,psi,lambda,var1,var2,covar,std1,std2,corr);
  }

  public static Scale parse(String string) {
    String[] cells = string.split("\\s+");
    if ( cells.length != 3 ) throw new IllegalArgumentException("Expecting three numbers.");
    double phi = Double.parseDouble(cells[0]);
    double psi = Double.parseDouble(cells[1]);
    double lambda = Double.parseDouble(cells[2]);
    return precision(phi, psi, lambda);
  }

  public String toString() {
    return String.format("%5.4f %5.4f %5.4f", phi, psi, lambda);
  }

}