/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP;

import HMMDP.BaseDistributions.Angles;
import HMMDP.BaseDistributions.BivariateVonMises;
import HMMDP.BaseDistributions.BivariateWishart;
import HMMDP.BaseDistributions.DistributionArray;
import HMMDP.BaseDistributions.Scale;
import java.util.Random;

/**
 *  Distribution for both mean and scale parameters, assuming that they are independent.
 */
public class HiddenMarkovModelFull implements DistributionArray<ParameterSet>{
//    private DistributionArray<Angles> centeringDistribution;
    private HiddenMarkovModelMean centeringDistribution;
    private IndepWishartChain precisionDistribution;
    private final int length;
    
    public HiddenMarkovModelFull(HiddenMarkovModelMean anglesDist, IndepWishartChain scaleDist) throws Exception{
       if(anglesDist.getLength()!=scaleDist.getLength()) throw new Exception("Component DistributionArrays must have identical lengths.");   
       centeringDistribution=anglesDist;
       precisionDistribution=scaleDist;
       length=scaleDist.getLength();
    }
    
    public double logDensity(ParameterSet[] x) throws Exception {        
        return logDensityKernel(x)+logNormalizingConstant();
    }
    
    public double logDensityKernelPrecision(ParameterSet[] x) throws Exception {
        if(x.length!=length) throw new Exception ("Incorrect ParameterSet[] dimension.");
        Scale[] scales=new Scale[length];
        for(int i=0;i<length;i++){
            scales[i]=x[i].scale;
        }
        return precisionDistribution.logDensity(scales);
    }

    
    public double logDensityKernelMean(ParameterSet[] x) throws Exception{
        if(x.length!=length) throw new Exception ("Incorrect ParameterSet[] dimension.");
        Angles[] centers=new Angles[length];
        for(int i=0;i<length;i++){
            centers[i]=x[i].center;
        }
        return centeringDistribution.logDensity(centers);
    }

    public double logDensityKernel(ParameterSet[] x) throws Exception{
        return logDensityKernelMean(x)+logDensityKernelPrecision(x);
    }
    
    public double logNormalizingConstantMean(){
        return centeringDistribution.logNormalizingConstant();
    }
    
    
    public double logNormalizingConstantPrecision(){
        return precisionDistribution.logNormalizingConstant();
    }
    
    public double logNormalizingConstant() {
        return logNormalizingConstantMean()+logNormalizingConstantPrecision();
    }
    
    
    // Exception will only come up if the component distributions have different dimensions.
    // Checked at initialization.
    public ParameterSet[] sample(Random random) {
        ParameterSet[] temp = new ParameterSet[length];
        Angles[] centers = centeringDistribution.sample(random);
        Scale[] scales = precisionDistribution.sample(random);
        for(int i=0;i<length;i++){
            // Note that no empty Angles exception will be thrown under these circumstances.
            try {
                temp[i] = new ParameterSet(centers[i], scales[i]);
            } catch (Exception ex) {}
        }
        return(temp);
    }
    
    public Angles[] sampleFullConditionalMeans(Angles[] theta, Angles[][] data, Scale[] scales, Random random) throws Exception{
        return centeringDistribution.sampleFullConditional(theta, data, scales, random);
    }

     public BivariateVonMises[] returnPriors(Angles[] theta, Random random) throws Exception{
        return centeringDistribution.returnPriors(theta, random);
    }
    
    public int getLength(){
        return length;
    }
    
    public BivariateWishart reportScale(int pos){
        return precisionDistribution.reportPosition(pos);
    }    

    
}
