/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP;

import HMMDP.BaseDistributions.Angles;
import HMMDP.BaseDistributions.BivariateVonMises;
import HMMDP.BaseDistributions.BivariateVonMisesMix;
import HMMDP.BaseDistributions.DistributionArray;
import HMMDP.BaseDistributions.Scale;
import java.util.Random;

/** This design will generate a hidden Markov Model using any random 
 * distribution of Angles. It's designed to work with the BivariateVonMises
 * and BivariateVonMises mix classes.
 */
public class HiddenMarkovModelMean implements DistributionArray<Angles> {

    private int length_;
    private int numStates_;
//    private Distribution<Angles>[] stateDists_;
    private BivariateVonMises[] stateDists_;
    private double[] initProb_;
    private double[][] transMatrix_;
    private double[] logInitProb_;
    private double[][] logTransMatrix_;
    /** Constructor for hidden Markov Model
      * length: length of Markov chain
      * stateDists: Parameter distributions for each state
      * initProb: probabilities for each state being the initial state
      * transMatrix: transition matrix for hidden markov model; In this parameterization, each row gives marginal
      *             probabilities
      * random: psuedo-random number generator
     */
    
    public HiddenMarkovModelMean(){    }
    
    public HiddenMarkovModelMean(int length, BivariateVonMises[] stateDists, double[] initProb, double[][] transMatrix) {
        length_ = length;
        stateDists_ = stateDists;
        numStates_ = stateDists.length;
        initProb_ = initProb;
        transMatrix_ = transMatrix;

        // log matrices for faster density calculation
        logInitProb_ = new double[numStates_];
        for (int i = 0; i < numStates_; i++) {
            logInitProb_[i] = Math.log(initProb_[i]);
        }

        logTransMatrix_ = new double[numStates_][numStates_];
        for (int i = 0; i < numStates_; i++) {
            for (int j = 0; j < numStates_; j++) {
                logTransMatrix_[i][j] = Math.log(transMatrix_[i][j]);
            }
        }
    }

    /** Method takes an array of probabilities and returns an integer with probability corresponding
      * to the value at that position in the array
     */
    private int randomInt(double[] probArray, Random random) {
        // Clone probability array. Allows us to ensure that entries sum to one. Find scaling factor, and check that all
        // values are > 0
        double[] tempProbs = (double[]) (probArray.clone());
        double total = 0.0;
        for (int i = 0; i < tempProbs.length; i++) {
            if (tempProbs[i] < 0) {
                throw new IllegalArgumentException("Elements of probability array must be non-negative.");
            }
            total += tempProbs[i];
        }
        // Rescale probabilities to sum to 1, and change from density cdf
        tempProbs[0] /= total;
        for (int i = 1; i < tempProbs.length; i++) {
            tempProbs[i] = tempProbs[i] / total + tempProbs[i - 1];
        }

        double value = random.nextDouble();
        int index = 0;
        while (tempProbs[index] <= value) {
            index++;
        }
        return (index);
    }

   
    /** Fast recursion density from Scott (JASA,2008)
      * Scales up at O((s^2)n) rather that O(s^n) of previous algorithm
     */
    public double logDensity(Angles[] x) {
        double[] temp = fastDensityRecurse( x, length_-1);
        double temp2 = 0.0;
        for(int i=0;i<numStates_;i++){
            temp2+=Math.exp(logInitProb_[i]+temp[i]);
        }
        return Math.log(temp2);
    }

    private double[] fastDensityRecurse(Angles[] x, int pos) {
        double[] temp= new double[numStates_];
        double[] next;
        
        if (pos>0) {
            for(int i=0;i<numStates_;i++) temp[i]=0.0;
            next = fastDensityRecurse(x,pos-1);
            for(int i=0;i<numStates_;i++){
                for(int j=0;j<numStates_;j++){
                    temp[i]+=Math.exp(logTransMatrix_[i][j]+next[j]);
                }             
            }
        } else 
            for(int i=0;i<numStates_;i++) temp[i]=1.0;
       
        for(int i=0;i<numStates_;i++){
            temp[i]*=Math.exp(stateDists_[i].logDensity(x[pos]));
            temp[i]=Math.log(temp[i]);
        }
        return temp;
    }
    
    public Angles[] sampleFullConditional(Angles[] x,Angles[][] data, Scale[] scales, Random random) throws Exception{
      // Algorithm from Chib 1996

              
      // Initialize posterior transition state matrix
      // Rows are positions, columns are states 
      double[][] postTrans = new double[length_][numStates_];
      // Assumes you give stationary state to logInitProb
      for(int i=0;i<numStates_;i++) postTrans[0][i] = Math.exp(logInitProb_[i]+stateDists_[i].logDensity(x[0]));

      // Initialize all other rows; i is sequence position, j is current state, k is next state
      for(int i=1;i<length_;i++){
          for(int j=0;j<numStates_;j++){
              // Initialize
              postTrans[i][j]=0;
            for(int k=0;k<numStates_;k++){
                // adding p(s_i-1=k|Y_i-1)p(s_i=j|s_i-1=k)
                postTrans[i][j]+=Math.exp(postTrans[i-1][k]+logTransMatrix_[k][j]);
            }
            // multiply times p(y_i|Y_i-1,state=j) Possibly should be full conditional, not prior
            postTrans[i][j]=(Math.log(postTrans[i][j])+stateDists_[j].logDensity(x[i]));
          }
      }
      
      // Sample from posterior dist of states
      int[] states = new int[length_];
      double[] tempProbs = new double[numStates_];
      // Initialize nth sequence position
      for(int i=0;i<numStates_;i++){
          tempProbs[i]=Math.exp(postTrans[length_-1][i]);
      }
      states[length_-1]=randomInt(tempProbs,random);
      // Initialize remainder of states
      for(int i=2;i<=length_;i++){
        for(int j=0;j<numStates_;j++){
          tempProbs[j]=Math.exp(postTrans[length_-i][j]+logTransMatrix_[j][states[length_-i+1]]);
        }
        states[length_-i]=randomInt(tempProbs,random);
      }
     // for(int i=0;i<length_;i++)
     //   System.out.print(states[i]);
     // System.out.println();
      
      // Generate new angles based on states, will eventually use posterior dist rather that 
      // prior
      Angles[] finalArray = new Angles[length_];
      for(int i=0;i<length_;i++){
          // Get data for ith positon
          Angles[] tempData = new Angles[data.length];
          for(int j=0;j<data.length;j++){
              tempData[j]=data[j][i];
          }
          finalArray[i]=(stateDists_[states[i]].fullConditional(tempData, scales[i])).sample(random);
      }
      
      return finalArray;
    }
    
    
    public BivariateVonMises[] returnPriors(Angles[] x, Random random) throws Exception{
      // Algorithm from Chib 1996
      // Initialize posterior transition state matrix
      // Rows are positions, columns are states 
      double[][] postTrans = new double[length_][numStates_];
      
      for(int i=0;i<numStates_;i++) postTrans[0][i] = Math.exp(logInitProb_[i]+stateDists_[i].logDensity(x[0]));

      // Initialize all other rows; i is sequence position, j is current state, k is next state
      for(int i=1;i<length_;i++){
          for(int j=0;j<numStates_;j++){
              // Initialize
              postTrans[i][j]=0;
            for(int k=0;k<numStates_;k++){
                // adding p(s_i-1=k|Y_i-1)p(s_i=j|s_i-1=k)
                postTrans[i][j]+=Math.exp(postTrans[i-1][k]+logTransMatrix_[k][j]);
            }
            // multiply times p(y_i|Y_i-1,state=j) Possibly should be full conditional, not prior
            postTrans[i][j]=(Math.log(postTrans[i][j])+stateDists_[j].logDensity(x[i]));
          }
      }
      
      // Sample from posterior dist of states
      int[] states = new int[length_];
      double[] tempProbs = new double[numStates_];
      // Initialize nth sequence position
      for(int i=0;i<numStates_;i++){
          tempProbs[i]=Math.exp(postTrans[length_-1][i]);
      }
      states[length_-1]=randomInt(tempProbs,random);
      // Initialize remainder of states
      for(int i=2;i<=length_;i++){
        for(int j=0;j<numStates_;j++){
          tempProbs[j]=Math.exp(postTrans[length_-i][j]+logTransMatrix_[j][states[length_-i+1]]);
        }
        states[length_-i]=randomInt(tempProbs,random);
      }

      // Return an array of proposal distributions based on the states.
      BivariateVonMises[] priorDists = new BivariateVonMises[length_];
      for(int i=0;i<length_;i++){
              priorDists[i]=stateDists_[states[i]];
      }
      
      return priorDists;   
    }
    
    /**
     * Returns a value proportional to the full conditional of parameter set x given the data.
     */
    public double logFullConditional(Angles[] x, Angles[][] data, Scale[] precisions) throws Exception{
        VonMisesChain condDist=new VonMisesChain(x,precisions);
        double likelihood = 0.0;
        return logDensity(x)+condDist.logDensity(data);
    }
    
     /**
      *  No normalizing constant, so same as logDensity.
      */
    public double logDensityKernel(Angles[] x) throws Exception {
        return logDensity(x);
    }

    /**
     *  The implementation of the density function doesn't use a normalizing constant.
     */
    public double logNormalizingConstant() {
        return 0.0;
    }

    public Angles[] sample(Random random) {

        // state array for debugging    
        int[] states = new int[length_];

        Angles[] obs = new Angles[length_];
        // Initialize first element using initProb matrix
        states[0] = randomInt(initProb_, random);
        obs[0] = (Angles) (stateDists_[states[0]].sample(random));
        // Using initial value and transition state matrix, initialize all subsequent elements
        for (int i = 1; i < length_; i++) {
            states[i] = randomInt(transMatrix_[states[i - 1]], random);
            obs[i] = (Angles) (stateDists_[states[i]].sample(random));
        }
        /* 
        // For debugging
        for(int i=0;i<states.length;i++){
        System.out.print(states[i]+" ");
        }
        System.out.print("\n");
         */
        return obs;
    }

    public int getLength() {
        return length_;
    }
}
 
        