/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP;

import HMMDP.BaseDistributions.Angles;
import HMMDP.BaseDistributions.BivariateVonMises;
import HMMDP.BaseDistributions.BivariateVonMisesMix;
import HMMDP.BaseDistributions.Scale;
import java.util.Random;

// Implements the peptide specific hidden Markov model with special priors for proline and glycine.
public class HiddenMarkovModelPeptide extends HiddenMarkovModelMean{

    private int length_;
    private int numStates_;
//    private Distribution<Angles>[] stateDists_;
    private BivariateVonMises[][] stateDists_;
    private double[] initProb_;
    private double[][] transMatrix_;
    private double[] logInitProb_;
    private double[][] logTransMatrix_;
    private int[] aminoAcids_;
    /** Constructor for hidden Markiv Model
      * length: length of Markov chain
      * stateDists: Parameter distributions for each state
      * initProb: probabilities for each state being the initial state
      * transMatrix: transition matrix for hidden markov model; In this parameterization, each row gives marginal
      *             probabilities
      * random: psuedo-random number generator
     */
    public HiddenMarkovModelPeptide(int length, double[] initProb, double[][] transMatrix, int[] aminoAcids) throws Exception {
        length_ = length;
        // Initialize array of state dist arrays
        // stateDists[0] = regular states
        // stateDists[1] = proline states
        // stateDists[2] = glycine states
        stateDists_ = new BivariateVonMises[3][];
        stateDists_[0] =generalInit();
        stateDists_[1] =proInit();
        stateDists_[2] =glyInit();

        // Check that the amino acid list is of the correct length.
        
        // Define the amino acid labels
        aminoAcids_= aminoAcids;
        // Check that dimension is correct.
        if(aminoAcids_.length!=length_){
            System.out.println("Amino acid string is of incorrect length.");
            System.exit(0);
        }
        numStates_ = stateDists_[0].length;

        initProb_ = initProb;
        transMatrix_ = transMatrix;

        // log matrices for faster density calculation
        logInitProb_ = new double[numStates_];
        for (int i = 0; i < numStates_; i++) {
            logInitProb_[i] = Math.log(initProb_[i]);
        }

        logTransMatrix_ = new double[numStates_][numStates_];
        for (int i = 0; i < numStates_; i++) {
            for (int j = 0; j < numStates_; j++) {
                logTransMatrix_[i][j] = Math.log(transMatrix_[i][j]);
            }
        }
    }

    /** Method takes an array of probabilities and returns an integer with probability corresponding
      * to the value at that position in the array
     */
    private int randomInt(double[] probArray, Random random) {
        // Clone probability array. Allows us to ensure that entries sum to one. Find scaling factor, and check that all
        // values are > 0
        double[] tempProbs = (double[]) (probArray.clone());
        double total = 0.0;
        for (int i = 0; i < tempProbs.length; i++) {
            if (tempProbs[i] < 0) {
                throw new IllegalArgumentException("Elements of probability array must be non-negative.");
            }
            total += tempProbs[i];
        }
        // Rescale probabilities to sum to 1, and change from density cdf
        tempProbs[0] /= total;
        for (int i = 1; i < tempProbs.length; i++) {
            tempProbs[i] = tempProbs[i] / total + tempProbs[i - 1];
        }

        double value = random.nextDouble();
        int index = 0;
        while (tempProbs[index] <= value) {
            index++;
        }
        return (index);
    }

   
    /** Fast recursion density from Scott (JASA,2008)
      * Scales up at O((s^2)n) rather that O(s^n) of previous algorithm
      * Slightly unstable. Could implement with logs.
     */
    @Override
    public double logDensity(Angles[] x) {
        double[] temp = fastDensityRecurse( x, length_-1);
        double temp2 = 0.0;
        for(int i=0;i<numStates_;i++){
            temp2+=Math.exp(logInitProb_[i]+temp[i]);
        }
        return Math.log(temp2);
    }

    private double[] fastDensityRecurse(Angles[] x, int pos) {
        double[] temp= new double[numStates_];
        double[] next;
        
        if (pos>0) {
            for(int i=0;i<numStates_;i++) temp[i]=0.0;
            next = fastDensityRecurse(x,pos-1);
            for(int i=0;i<numStates_;i++){
                for(int j=0;j<numStates_;j++){
                    temp[i]+=Math.exp(logTransMatrix_[i][j]+next[j]);
                }             
            }
        } else 
            for(int i=0;i<numStates_;i++) temp[i]=1.0;
       
        for(int i=0;i<numStates_;i++){
            temp[i]*=Math.exp(stateDists_[aminoAcids_[pos]][i].logDensity(x[pos]));
            temp[i]=Math.log(temp[i]);
        }
        return temp;
    }


    @Override
    public BivariateVonMises[] returnPriors(Angles[] x, Random random){
        // Initialize posterior transition state matrix
      // Rows are positions, columns are states
      double[][] postTrans = new double[length_][numStates_];
    
      for(int i=0;i<numStates_;i++) postTrans[0][i] = Math.exp(logInitProb_[i]+stateDists_[aminoAcids_[0]][i].logDensity(x[0]));

      // Initialize all other rows; i is sequence position, j is current state, k is next state
      for(int i=1;i<length_;i++){
          for(int j=0;j<numStates_;j++){
              // Initialize
              postTrans[i][j]=0;
            for(int k=0;k<numStates_;k++){
                // adding p(s_i-1=k|Y_i-1)p(s_i=j|s_i-1=k)
                postTrans[i][j]+=Math.exp(postTrans[i-1][k]+logTransMatrix_[k][j]);
            }
            // multiply times p(y_i|Y_i-1,state=j) Possibly should be full conditional, not prior
            postTrans[i][j]=(Math.log(postTrans[i][j])+stateDists_[aminoAcids_[i]][j].logDensity(x[i]));
          }
      }

      // Sample from posterior dist of states
      int[] states = new int[length_];
      double[] tempProbs = new double[numStates_];
      // Initialize nth sequence position
      for(int i=0;i<numStates_;i++){
          tempProbs[i]=Math.exp(postTrans[length_-1][i]);
      }
      states[length_-1]=randomInt(tempProbs,random);
      // Initialize remainder of states
      for(int i=2;i<=length_;i++){
        for(int j=0;j<numStates_;j++){
          tempProbs[j]=Math.exp(postTrans[length_-i][j]+logTransMatrix_[j][states[length_-i+1]]);
        }
        states[length_-i]=randomInt(tempProbs,random);
      }
     // for(int i=0;i<length_;i++)
     //   System.out.print(states[i]);
     // System.out.println();

      // Generate new angles based on states, will eventually use posterior dist rather that
      // prior
      BivariateVonMises[] tempPriors = new BivariateVonMises[length_];
      for(int i=0;i<length_;i++){
          tempPriors[i]=stateDists_[aminoAcids_[i]][states[i]];
      }

      return  tempPriors;

    }

    // Uses incorecct full conditional for means. Use only as a proposal.
    @Override
    public Angles[] sampleFullConditional(Angles[] x,Angles[][] data, Scale[] scales, Random random) throws Exception{
      // Algorithm from Chib 1996

              
      // Initialize posterior transition state matrix
      // Rows are positions, columns are states 
      double[][] postTrans = new double[length_][numStates_];
      // FAKE FIRST ROW: NEED STATIONARY STATE TO REPLACE LOGINITPROB?
      for(int i=0;i<numStates_;i++) postTrans[0][i] = Math.exp(logInitProb_[i]+stateDists_[aminoAcids_[0]][i].logDensity(x[0]));

      // Initialize all other rows; i is sequence position, j is current state, k is next state
      for(int i=1;i<length_;i++){
          for(int j=0;j<numStates_;j++){
              // Initialize
              postTrans[i][j]=0;
            for(int k=0;k<numStates_;k++){
                // adding p(s_i-1=k|Y_i-1)p(s_i=j|s_i-1=k)
                postTrans[i][j]+=Math.exp(postTrans[i-1][k]+logTransMatrix_[k][j]);
            }
            // multiply times p(y_i|Y_i-1,state=j) Possibly should be full conditional, not prior
            postTrans[i][j]=(Math.log(postTrans[i][j])+stateDists_[aminoAcids_[i]][j].logDensity(x[i]));
          }
      }
      
      // Sample from posterior dist of states
      int[] states = new int[length_];
      double[] tempProbs = new double[numStates_];
      // Initialize nth sequence position
      for(int i=0;i<numStates_;i++){
          tempProbs[i]=Math.exp(postTrans[length_-1][i]);
      }
      states[length_-1]=randomInt(tempProbs,random);
      // Initialize remainder of states
      for(int i=2;i<=length_;i++){
        for(int j=0;j<numStates_;j++){
          tempProbs[j]=Math.exp(postTrans[length_-i][j]+logTransMatrix_[j][states[length_-i+1]]);
        }
        states[length_-i]=randomInt(tempProbs,random);
      }
     // for(int i=0;i<length_;i++)
     //   System.out.print(states[i]);
     // System.out.println();
      
      // Generate new angles based on states, will eventually use posterior dist rather that 
      // prior
      Angles[] finalArray = new Angles[length_];
      for(int i=0;i<length_;i++){
          // Get data for ith positon
          Angles[] tempData = new Angles[data.length];
          for(int j=0;j<data.length;j++){
              tempData[j]=data[j][i];
          }
          finalArray[i]=(stateDists_[aminoAcids_[i]][states[i]].fullConditional(tempData, scales[i])).sample(random);
      }
      
      return finalArray;
    }
    
    /**
     * Returns a value proportional to the full conditional of parameter ser x given the data.
     */
    @Override
    public double logFullConditional(Angles[] x, Angles[][] data, Scale[] precisions) throws Exception{
        VonMisesChain condDist=new VonMisesChain(x,precisions);
        double likelihood = 0.0;
        return logDensity(x)+condDist.logDensity(data);
    }
    
     /**
      *  No normalizing constant, so same as logDensity.
      */
    public double logDensityKernel(Angles[] x) throws Exception {
        return logDensity(x);
    }

    /**
     *  The implementation of the density function doesn't use a normalizing constant.
     */
    @Override
    public double logNormalizingConstant() {
        return 0.0;
    }

    @Override
    public Angles[] sample(Random random) {

        // state array for debugging    
        int[] states = new int[length_];

        Angles[] obs = new Angles[length_];
        // Initialize first element using initProb matrix
        states[0] = randomInt(initProb_, random);
        obs[0] = (Angles) (stateDists_[aminoAcids_[0]][states[0]].sample(random));
        // Using initial value and transition state matrix, initialize all subsequent elements
        for (int i = 1; i < length_; i++) {
            states[i] = randomInt(transMatrix_[states[i - 1]], random);
            obs[i] = (Angles) (stateDists_[aminoAcids_[i]][states[i]].sample(random));
        }
        /* 
        // For debugging
        for(int i=0;i<states.length;i++){
        System.out.print(states[i]+" ");
        }
        System.out.print("\n");
         */
        return obs;
    }

    @Override
    public int getLength() {
        return length_;
    }
    
    public static BivariateVonMisesMix[] generalInit() throws Exception{
        // Set up informative priors
        // Turn
        BivariateVonMises turn1 = new BivariateVonMises(new Angles(-1,2.5),Scale.precision(21.33,21.33,10.67));
        BivariateVonMises turn2 = new BivariateVonMises(new Angles(-1.2,-.2),Scale.precision(8.33,8.33,4.17));
        BivariateVonMises turn3 = new BivariateVonMises(new Angles(1,.6),Scale.precision(33.33,8.33,8.33));
        BivariateVonMises[] turnSet = {turn1,turn2,turn3};
        double[] turnWeights = {.1,.8,.1};
        BivariateVonMisesMix turnPrior = new BivariateVonMisesMix(turnSet, turnWeights);
        
        // Coil
        BivariateVonMises coil1 = new BivariateVonMises(new Angles(-2,2.5),Scale.precision(4,4,0));
        BivariateVonMises coil2 = new BivariateVonMises(new Angles(-1,2.5),Scale.precision(21.33,21.33,10.67));
        BivariateVonMises coil3 = new BivariateVonMises(new Angles(-2,0),Scale.precision(6.25,6.25,0));
        BivariateVonMises coil4 = new BivariateVonMises(new Angles(1,1),Scale.precision(12.21,12.21,3.66));
        BivariateVonMises[] coilSet = {coil1,coil2,coil3,coil4};
        double[] coilWeights = {.625,.208,.125,.043};
        BivariateVonMisesMix coilPrior = new BivariateVonMisesMix(coilSet, coilWeights);
        
        //Helix
        BivariateVonMises helix1 = new BivariateVonMises(new Angles(-1,-.5),Scale.precision(21.33,21.33,10.67));
        BivariateVonMises[] helixSet = {helix1};
        double[] helixWeights = {1};
        BivariateVonMisesMix helixPrior = new BivariateVonMisesMix(helixSet, helixWeights);
        
        //Sheet
        BivariateVonMises sheet1 = new BivariateVonMises(new Angles(-2,2.5),Scale.precision(5.33,21.33,5.33));
        BivariateVonMises[] sheetSet = {sheet1};
        double[] sheetWeights = {1};
        BivariateVonMisesMix sheetPrior = new BivariateVonMisesMix(sheetSet, sheetWeights);
        
        return new BivariateVonMisesMix[] {turnPrior, coilPrior, helixPrior,sheetPrior};
       
    }
    
    private BivariateVonMisesMix[] proInit() throws Exception{
        // Set up informative priors
        // Turn
        BivariateVonMises turn1 = new BivariateVonMises(new Angles(-1,2.5),Scale.precision(21.33,21.33,10.67));
        BivariateVonMises turn2 = new BivariateVonMises(new Angles(-1.2,-.5),Scale.precision(14.81,14.81,7.41));
        BivariateVonMises[] turnSet = {turn1,turn2};
        double[] turnWeights = {.2,.8};
        BivariateVonMisesMix turnPrior = new BivariateVonMisesMix(turnSet, turnWeights);
        
        // Coil
        BivariateVonMises coil1 = new BivariateVonMises(new Angles(-1.2,2.5),Scale.precision(17.36,9.77,7.81));
        BivariateVonMises coil2 = new BivariateVonMises(new Angles(-1.2,-0.5),Scale.precision(29.54,17.36,13.59));
        BivariateVonMises coil3 = new BivariateVonMises(new Angles(-1.35,1.0),Scale.precision(11.11,4.0,0));
        BivariateVonMises[] coilSet = {coil1,coil2,coil3};
        double[] coilWeights = {.789,.105,.105};
        BivariateVonMisesMix coilPrior = new BivariateVonMisesMix(coilSet, coilWeights);
        
        //Helix
        BivariateVonMises helix1 = new BivariateVonMises(new Angles(-1,-.6),Scale.precision(29.54,17.36,13.59));
        BivariateVonMises[] helixSet = {helix1};
        double[] helixWeights = {1};
        BivariateVonMisesMix helixPrior = new BivariateVonMisesMix(helixSet, helixWeights);
        
        //Sheet
        BivariateVonMises sheet1 = new BivariateVonMises(new Angles(-1.2,2.5),Scale.precision(17.58,6.87,3.3));
        BivariateVonMises sheet2 = new BivariateVonMises(new Angles(-1.2,-.5),Scale.precision(29.54,17.36,13.59));
        BivariateVonMises sheet3 = new BivariateVonMises(new Angles(-1.35,1),Scale.precision(25,4,0));
        BivariateVonMises[] sheetSet = {sheet1,sheet2,sheet3};
        double[] sheetWeights = {0.56,0.32,0.12};
        BivariateVonMisesMix sheetPrior = new BivariateVonMisesMix(sheetSet, sheetWeights);
        
        return new BivariateVonMisesMix[] {turnPrior, coilPrior, helixPrior,sheetPrior};
       
    }
        
    private BivariateVonMisesMix[] glyInit() throws Exception{
        // Set up informative priors
        // Turn
        BivariateVonMises turn1 = new BivariateVonMises(new Angles(1.5,.1),Scale.precision(12.76,12.76,7.65));
        BivariateVonMises turn2 = new BivariateVonMises(new Angles(-1.5,-.2),Scale.precision(29.54,17.36,13.59));
        BivariateVonMises turn3 = new BivariateVonMises(new Angles(1,-2.2),Scale.precision(11.11,4,0));
        BivariateVonMises[] turnSet = {turn1,turn2,turn3};
        double[] turnWeights = {.75,.1,.15};
        BivariateVonMisesMix turnPrior = new BivariateVonMisesMix(turnSet, turnWeights);
        
        // Coil
        BivariateVonMises coil1 = new BivariateVonMises(new Angles(-3,3),Scale.precision(2.78,6.25,0));
        BivariateVonMises coil2 = new BivariateVonMises(new Angles(1.5,.2),Scale.precision(17.36,17.36,10.42));
        BivariateVonMises coil3 = new BivariateVonMises(new Angles(1.8,3),Scale.precision(10.88,3.7,3.17));
        BivariateVonMises coil4 = new BivariateVonMises(new Angles(-1.5,.3),Scale.precision(13.72,11.11,0));
        BivariateVonMises coil5 = new BivariateVonMises(new Angles(-1.5,3),Scale.precision(10.88,3.7,3.17));
        BivariateVonMises[] coilSet = {coil1,coil2,coil3,coil4,coil5};
        double[] coilWeights = {.29,.11,.27,.07,.27};
        BivariateVonMisesMix coilPrior = new BivariateVonMisesMix(coilSet, coilWeights);
        
        //Helix
        BivariateVonMises helix1 = new BivariateVonMises(new Angles(-1,-.6),Scale.precision(29.54,17.36,13.59));
        BivariateVonMises[] helixSet = {helix1};
        double[] helixWeights = {1};
        BivariateVonMisesMix helixPrior = new BivariateVonMisesMix(helixSet, helixWeights);
        
        //Sheet
        BivariateVonMises sheet1 = new BivariateVonMises(new Angles(-2.7,3),Scale.precision(5.33,8.33,3.33));
        BivariateVonMises sheet2 = new BivariateVonMises(new Angles(1.5,.2),Scale.precision(17.36,17.36,10.42));
        BivariateVonMises sheet3 = new BivariateVonMises(new Angles(1.8,-3),Scale.precision(8.5,4.17,1.19));
        BivariateVonMises sheet4 = new BivariateVonMises(new Angles(-1.4,-.4),Scale.precision(21.43,17.36,11.57));
        BivariateVonMises sheet5 = new BivariateVonMises(new Angles(-1.5,3),Scale.precision(11.11,6.25,0));
        BivariateVonMises[] sheetSet = {sheet1,sheet2,sheet3,sheet4,sheet5};
        double[] sheetWeights = {.31,.13,.26,.05,.26};
        BivariateVonMisesMix sheetPrior = new BivariateVonMisesMix(sheetSet, sheetWeights);
        
        return new BivariateVonMisesMix[] {turnPrior, coilPrior, helixPrior,sheetPrior};
       
    }
}
