/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP;

import HMMDP.BaseDistributions.Angles;
import HMMDP.BaseDistributions.BivariateVonMises;
import HMMDP.BaseDistributions.BivariateWishart;
import HMMDP.BaseDistributions.Scale;
import java.util.Random;
import java.text.DecimalFormat;

public class HmmClustering {

    // For both data and parameters, each row is a single observation
    private Angles[][] data;
    private double mass;
    private ParameterSet[][] parameters;
    private VonMisesChain[] distributions;
    private HiddenMarkovModelFull prior;
    // Current iteration, 0 at initialization
    private int clusterIteration = 0;
    private int paramIteration=0;
    // Measure acceptance prob across all positions for all iterations
    private double totalMeanAccept = 0;
    private double totalScaleAccept=0;
    private double currentMeanAccept=0;
    private double currentScaleAccept=0;
     
    private Random random;
    // Number of unique clusters
    private int unique;
    // Gives current cluster label of each observation
    private int[] cluster;
    // Gives sive of every unique cluster
    private int[] clusterSize;
     
    /**
     * Constructor that has a flag to initially set all observations to be in the same cluster,
     * or all in different clusters
     */
    public HmmClustering(Angles[][] data, double mass, HiddenMarkovModelFull prior, Random random, boolean oneCluster) throws Exception {
        if (data[0].length != prior.getLength()) {
            throw new Exception("Invalid prior. Incorrect dimension.");
        }
        this.data = data;
        this.mass = mass;
        this.prior = prior;
        this.parameters = new ParameterSet[data.length][data[0].length];
        this.distributions = new VonMisesChain[data.length];
        this.cluster = new int[data.length];

        this.random = random;
        if (oneCluster) {
            unique = 1;
            ParameterSet[] temp = this.prior.sample(random);
            VonMisesChain tempDistribution = new VonMisesChain(temp);
            for (int i = 0; i < distributions.length; i++) {
                parameters[i] = temp;
                distributions[i] = tempDistribution;
                cluster[i] = 0;
            }
        } else {
            unique = data.length;
            for (int i = 0; i < distributions.length; i++) {
                parameters[i] = this.prior.sample(random);
                distributions[i] = new VonMisesChain(parameters[i]);
                cluster[i] = i;
            }
        }
        countUnique();
    }

    /**
     * By default, initializes with each observation in a separate cluster
     */
    public HmmClustering(Angles[][] data, double mass, HiddenMarkovModelFull prior, Random random) throws Exception {
        if (data[0].length != prior.getLength()) {
            throw new Exception("Invalid prior. Incorrect dimension.");
        }
        this.data = data;
        this.mass = mass;
        this.prior = prior;
        this.parameters = new ParameterSet[data.length][data[0].length];
        this.distributions = new VonMisesChain[data.length];
        this.random = random;
        this.cluster = new int[data.length];

        unique = data.length;
        for (int i = 0; i < distributions.length; i++) {
            parameters[i] = this.prior.sample(random);
            distributions[i] = new VonMisesChain(parameters[i]);
            cluster[i] = i;
        }
        countUnique();
    }

    /**
     *Updates with Auxillary Gibbs sampling of Neal2000 w/ m=1
     */
    public void update() throws Exception {
        //updateClusters();
        // Faster Aux Gibbs Sampler.
        updateClustersNew();
        for(int i=0;i<1;i++){
         //Use full conditional for means
        // INCORRECT!
      //  updateParamsNew();
        // CORRECTED
        updateParamsFixed();
        //Use indep sampler for means
     //     updateParams();
        }

        
    }

    public void report(){
      // Calculate entropy and largest cluster size
      double entropy=0.0;
      double max = 0.0;
      double total = (double)data.length;
      for(int i=0;i<clusterSize.length;i++){
          entropy+=clusterSize[i]/total*Math.log(clusterSize[i]/total);
          if(clusterSize[i]>max)
              max=(double)clusterSize[i];
      }
      entropy*=(-1);
      max/=total;
      DecimalFormat twoPlaces = new DecimalFormat("0.00");
      System.out.print(clusterIteration+" "+unique+" "+twoPlaces.format(currentMeanAccept)+" "+twoPlaces.format(totalMeanAccept/clusterIteration)+" "
              +twoPlaces.format(currentScaleAccept)+" "+twoPlaces.format(totalScaleAccept/clusterIteration));
 //  Prints sizes of all clusters     
 //     for(int i=0;i<unique;i++){
 //          System.out.print(" "+clusterSize[i]);
 //      }
      // Print entropy and percentage obs in max cluster.
      System.out.print(" "+twoPlaces.format(entropy)+" "+twoPlaces.format(max));
       System.out.println();
    }
    
    private void updateParams() throws Exception{
            int m = parameters.length;
        double meanAccept=0;
        double scaleAccept=0;
        // Update cluster params based on posterior distributions.
        for(int i=0;i<unique;i++){
            
            // Generate data set for each cluster.
            Angles[][] clusterData=new Angles[clusterSize[i]][];
            int current=0;
            int example=0;
            for(int j=0;j<m;j++){
               if(cluster[j]==i){
                    clusterData[current]=data[j];
                    if(current==0) example=j; 
                    current++;
                    
                }
            }

            // Update each position
            for(int j=0;j<clusterData[0].length;j++){
              Angles[] posData = new Angles[clusterData.length];
              for(int k=0;k<clusterData.length;k++){
                  posData[k]=clusterData[k][j];
              }
              
              
              //Old  mean parameter update, used to udate mean param and scale param for each pos
              BivariateVonMises meanProposalDist = BivariateVonMises.fullConditionalUniform(posData, parameters[example][j].scale);
              // Calculate likelihood of data under old distribution
              double oldPosterior = ((BivariateVonMises)distributions[example].reportPosition(j)).logLikelihoodKernel(posData);
              oldPosterior+=prior.logDensityKernelMean(parameters[example]);
              ParameterSet oldParams = new ParameterSet(parameters[example][j].center,parameters[example][j].scale);
          
              // Propose new mean parameters
              parameters[example][j]=new ParameterSet(meanProposalDist.sample(random),parameters[example][j].scale);
              distributions[example].replacePosition(j, new BivariateVonMises(parameters[example][j]));
              double newPosterior=((BivariateVonMises)distributions[example].reportPosition(j)).logLikelihoodKernel(posData);
              newPosterior+=prior.logDensityKernelMean(parameters[example]);
              
              double ratio = (newPosterior - oldPosterior) +(meanProposalDist.logDensity(oldParams.center)-meanProposalDist.logDensity(parameters[example][j].center));
              if(Math.log(random.nextDouble())<ratio){
                 meanAccept++;
              }else{
                  parameters[example][j]=oldParams;
                  distributions[example].replacePosition(j, new BivariateVonMises(oldParams));
              }
              
              
              // Scale parameter update
              BivariateWishart scaleProposalDist=(prior.reportScale(j)).fullConditional(posData, parameters[example][j].center);
              oldPosterior = ((BivariateVonMises)distributions[example].reportPosition(j)).logLikelihood(posData);
              oldPosterior+=prior.reportScale(j).logDensity(parameters[example][j].scale);
              oldParams = new ParameterSet(parameters[example][j].center,parameters[example][j].scale);
             
              parameters[example][j]=new ParameterSet(parameters[example][j].center,scaleProposalDist.sample(random));
              distributions[example].replacePosition(j, new BivariateVonMises(parameters[example][j]));
              newPosterior = ((BivariateVonMises)distributions[example].reportPosition(j)).logLikelihood(posData);
              newPosterior+=prior.reportScale(j).logDensity(parameters[example][j].scale);

              ratio = (newPosterior - oldPosterior) +(scaleProposalDist.logDensity(oldParams.scale)-scaleProposalDist.logDensity(parameters[example][j].scale));
              if(Math.log(random.nextDouble())<ratio){
                  scaleAccept++;
              }else{
                  parameters[example][j]=oldParams;
                  distributions[example].replacePosition(j, new BivariateVonMises(oldParams));
              }
              
              
            }
            
        }
        paramIteration++;
        currentMeanAccept=meanAccept/parameters[0].length/unique;
        totalMeanAccept+=currentMeanAccept;
        currentScaleAccept=scaleAccept/parameters[0].length/unique;
        totalScaleAccept+=currentScaleAccept;

    }
    
    
    // Uses prior as proposal distribution
    private void updateParamsAlt() throws Exception{
        int m = parameters.length;
        double meanAccept=0;
        double scaleAccept=0;
        // Update cluster params based on posterior distributions.
        for(int i=0;i<unique;i++){
            
            // Generate data set for each cluster.
            Angles[][] clusterData=new Angles[clusterSize[i]][];
            int current=0;
            int example=0;
            for(int j=0;j<m;j++){
               // If observation is in the current cluster
               if(cluster[j]==i){
                    // add observation to cluster data set
                    clusterData[current]=data[j];
                    // find an example of the current cluster
                    if(current==0) example=j; 
                    current++;
                    
                }
            }
            
            // New mean update: update all mean params
            Angles[] means = new Angles[clusterData[0].length];
            Scale[] scales = new Scale[clusterData[0].length];
            for(int j=0;j<means.length;j++){
                means[j]=parameters[example][j].center;
                scales[j]=parameters[example][j].scale;
            }
            // Update the state of the Markov state chain
            BivariateVonMises[] currPriors = prior.returnPriors(means, random);
                        
            // Update each position - Means
            for(int j=0;j<clusterData[0].length;j++){
              Angles[] posData = new Angles[clusterData.length];
              for(int k=0;k<clusterData.length;k++){
                  posData[k]=clusterData[k][j];
              }
              // Mean parameter update
              BivariateVonMises meanProposalDist=currPriors[j];
              double oldPosterior = currPriors[j].logDensity(parameters[example][j].center);
              oldPosterior+=((BivariateVonMises)distributions[example].reportPosition(j)).logLikelihood(posData);
              
              ParameterSet proposalParams = new ParameterSet(meanProposalDist.sample(random),parameters[example][j].scale);
              BivariateVonMises proposalDist = new BivariateVonMises(proposalParams);
              
              double newPosterior = currPriors[j].logDensity(proposalParams.center);
              newPosterior+=proposalDist.logLikelihood(posData);
            
              
              
              double ratio = (newPosterior - oldPosterior) +(meanProposalDist.logDensity(parameters[example][j].center)-meanProposalDist.logDensity(proposalParams.center));
              if(Math.log(random.nextDouble())<ratio){
                  meanAccept++;
                  parameters[example][j]=proposalParams;
                  distributions[example].replacePosition(j, proposalDist);
              }
            }
                        
            // Update each position - scale
            for(int j=0;j<clusterData[0].length;j++){
              Angles[] posData = new Angles[clusterData.length];
              for(int k=0;k<clusterData.length;k++){
                  posData[k]=clusterData[k][j];
              }
              // Scale parameter update
              BivariateWishart scaleProposalDist=(prior.reportScale(j)).fullConditional(posData, parameters[example][j].center);
              double oldPosterior = prior.reportScale(j).logDensity(parameters[example][j].scale);
              oldPosterior+=((BivariateVonMises)distributions[example].reportPosition(j)).logLikelihood(posData);
              
              
              ParameterSet proposalParams = new ParameterSet(parameters[example][j].center, scaleProposalDist.sample(random));
              BivariateVonMises proposalDist = new BivariateVonMises(proposalParams);
              
              double newPosterior = prior.reportScale(j).logDensity(proposalParams.scale);
              newPosterior+=proposalDist.logLikelihood(posData);
            
              
              
              double ratio = (newPosterior - oldPosterior) +(scaleProposalDist.logDensity(parameters[example][j].scale)-scaleProposalDist.logDensity(proposalParams.scale));
              if(Math.log(random.nextDouble())<ratio){
                  scaleAccept++;
                  parameters[example][j]=proposalParams;
                  distributions[example].replacePosition(j, proposalDist);
              }
            }
            
        }
        paramIteration++;
        currentMeanAccept=meanAccept/parameters[0].length/unique;
        totalMeanAccept+=currentMeanAccept;
        currentScaleAccept=scaleAccept/parameters[0].length/unique;
        totalScaleAccept+=currentScaleAccept;
    }
    
    
    // Uses dist from JASA paper as proposal distribution
    private void updateParamsFixed() throws Exception{
        int m = parameters.length;
        double meanAccept=0;
        double scaleAccept=0;
        // Update cluster params based on posterior distributions.
        for(int i=0;i<unique;i++){
            
            // Generate data set for each cluster.
            Angles[][] clusterData=new Angles[clusterSize[i]][];
            int current=0;
            int example=0;
            for(int j=0;j<m;j++){
               // If observation is in the current cluster
               if(cluster[j]==i){
                    // add observation to cluster data set
                    clusterData[current]=data[j];
                    // find an example of the current cluster
                    if(current==0) example=j; 
                    current++;
                    
                }
            }
            
            // New mean update: update all mean params
            Angles[] means = new Angles[clusterData[0].length];
            Scale[] scales = new Scale[clusterData[0].length];
            for(int j=0;j<means.length;j++){
                means[j]=parameters[example][j].center;
                scales[j]=parameters[example][j].scale;
            }
            // Update the state of the Markov state chain
            BivariateVonMises[] currPriors = prior.returnPriors(means, random);
                        
            // Update each position - Means
            for(int j=0;j<clusterData[0].length;j++){
              Angles[] posData = new Angles[clusterData.length];
              for(int k=0;k<clusterData.length;k++){
                  posData[k]=clusterData[k][j];
              }
              // Mean parameter update
              // Note that the fullConditional function does not provide the true full conditional distribution
              BivariateVonMises meanProposalDist=currPriors[j].fullConditional(posData, parameters[example][j].scale);
              double oldPosterior = currPriors[j].logDensity(parameters[example][j].center);
              oldPosterior+=((BivariateVonMises)distributions[example].reportPosition(j)).logLikelihood(posData);
              
              ParameterSet proposalParams = new ParameterSet(meanProposalDist.sample(random),parameters[example][j].scale);
              BivariateVonMises proposalDist = new BivariateVonMises(proposalParams);
              
              double newPosterior = currPriors[j].logDensity(proposalParams.center);
              newPosterior+=proposalDist.logLikelihood(posData);
            
              
              
              double ratio = (newPosterior - oldPosterior) +(meanProposalDist.logDensity(parameters[example][j].center)-meanProposalDist.logDensity(proposalParams.center));
              if(Math.log(random.nextDouble())<ratio){
                  meanAccept++;
                  parameters[example][j]=proposalParams;
                  distributions[example].replacePosition(j, proposalDist);
              }
            }
                        
            // Update each position - scale
            for(int j=0;j<clusterData[0].length;j++){
              // Obtain cluster observations for the jth sequence position
              Angles[] posData = new Angles[clusterData.length];
              for(int k=0;k<clusterData.length;k++){
                  posData[k]=clusterData[k][j];
              }

              //System.out.println(j+" "+clusterData.length+" "+posData.length);

              BivariateWishart scaleProposalDist=(prior.reportScale(j)).fullConditional(posData, parameters[example][j].center);
              double oldPosterior = prior.reportScale(j).logDensity(parameters[example][j].scale);
              oldPosterior+=((BivariateVonMises)distributions[example].reportPosition(j)).logLikelihood(posData);
              
              
              ParameterSet proposalParams = new ParameterSet(parameters[example][j].center, scaleProposalDist.sample(random));
              BivariateVonMises proposalDist = new BivariateVonMises(proposalParams);
              
              double newPosterior = prior.reportScale(j).logDensity(proposalParams.scale);
              newPosterior+=proposalDist.logLikelihood(posData);
            
              
              
              double ratio = (newPosterior - oldPosterior) +(scaleProposalDist.logDensity(parameters[example][j].scale)-scaleProposalDist.logDensity(proposalParams.scale));
              if(Math.log(random.nextDouble())<ratio){
                  scaleAccept++;
                  parameters[example][j]=proposalParams;
                  distributions[example].replacePosition(j, proposalDist);
              }
            }
            
        }
        paramIteration++;
        currentMeanAccept=meanAccept/parameters[0].length/unique;
        totalMeanAccept+=currentMeanAccept;
        currentScaleAccept=scaleAccept/parameters[0].length/unique;
        totalScaleAccept+=currentScaleAccept;
    }
    
    
    
// BROKEN! DO NOT USE! The mean update is incorrect. Use updateParamsFixed instead.    
    private void updateParamsNew() throws Exception{
        int m = parameters.length;
        double meanAccept=0;
        double scaleAccept=0;
        // Update cluster params based on posterior distributions.
        for(int i=0;i<unique;i++){
            
            // Generate data set for each cluster.
            Angles[][] clusterData=new Angles[clusterSize[i]][];
            int current=0;
            int example=0;
            for(int j=0;j<m;j++){
               // If observation is in the current cluster
               if(cluster[j]==i){
                    // add observation to cluster data set
                    clusterData[current]=data[j];
                    // find an example of the current cluster
                    if(current==0) example=j; 
                    current++;
                    
                }
            }
            
            // New mean update: update all mean params
            Angles[] means = new Angles[clusterData[0].length];
            Scale[] scales = new Scale[clusterData[0].length];
            for(int j=0;j<means.length;j++){
                means[j]=parameters[example][j].center;
                scales[j]=parameters[example][j].scale;
            }
            
            means=prior.sampleFullConditionalMeans(means, clusterData, scales, random);
            
            for(int j=0;j<means.length;j++){
                parameters[example][j]=new ParameterSet(means[j],scales[j]);
                distributions[example].replacePosition(j, new BivariateVonMises(parameters[example][j]));
            }
            
            
            
            // Update each position
            for(int j=0;j<clusterData[0].length;j++){
              Angles[] posData = new Angles[clusterData.length];
              for(int k=0;k<clusterData.length;k++){
                  posData[k]=clusterData[k][j];
              }
              // Scale parameter update
              BivariateWishart scaleProposalDist=(prior.reportScale(j)).fullConditional(posData, parameters[example][j].center);
              double oldPosterior = prior.reportScale(j).logDensity(parameters[example][j].scale);
              oldPosterior+=((BivariateVonMises)distributions[example].reportPosition(j)).logLikelihood(posData);
              
              
              ParameterSet proposalParams = new ParameterSet(parameters[example][j].center, scaleProposalDist.sample(random));
              BivariateVonMises proposalDist = new BivariateVonMises(proposalParams);
              
              double newPosterior = prior.reportScale(j).logDensity(proposalParams.scale);
              newPosterior+=proposalDist.logLikelihood(posData);
            
              
              
              double ratio = (newPosterior - oldPosterior) +(scaleProposalDist.logDensity(parameters[example][j].scale)-scaleProposalDist.logDensity(proposalParams.scale));
              if(Math.log(random.nextDouble())<ratio){
                  scaleAccept++;
                  parameters[example][j]=proposalParams;
                  distributions[example].replacePosition(j, proposalDist);
              }
            }
            
        }
        paramIteration++;
        currentMeanAccept=1;
        totalMeanAccept+=currentMeanAccept;
        currentScaleAccept=scaleAccept/parameters[0].length/unique;
        totalScaleAccept+=currentScaleAccept;
    }

    // Slower Auxillary Gibbs sampler, but allows addition of multiple clusters
    // during a single iteration.
    private void updateClusters() throws Exception{
        int m = parameters.length;
        ParameterSet[] proposalParam = prior.sample(random);
        VonMisesChain proposalDist = new VonMisesChain(proposalParam);
        // Update clustering
        for (int i = 0; i < m; i++) {
            Angles[] currentData = data[i];
            double[] tempProbs = new double[m + 1];
            for (int j = 0; j < m; j++) {
                tempProbs[j] = Math.exp(distributions[j].logDensity(currentData));
            }
            tempProbs[i] = 0;
            tempProbs[m] = mass * Math.exp(proposalDist.logDensity(currentData));
            int index = randomInt(tempProbs,random);
          //  System.out.print(index+" ");
            if(index==m){
                parameters[i]=proposalParam;
                distributions[i]=proposalDist;
            } else {
                distributions[i]=distributions[index];
                parameters[i]=parameters[index];
            }
            proposalParam = prior.sample(random);
            proposalDist =  new VonMisesChain(proposalParam);
        }

        // Updates cluster array and unique cluster count.
        countUnique();
        clusterIteration++;
    }



    // Faster Auxillary Gibbs algorithm. Now also includes improved mixing provisions.
    private void updateClustersNew() throws Exception{
        int m = parameters.length;
        //Find examples of each cluster type.
        ParameterSet[][] clusterParams = new ParameterSet[unique+1][];
        VonMisesChain[] clusterDists = new VonMisesChain[unique+1];
        clusterParams[unique]=prior.sample(random);
        clusterDists[unique]=new VonMisesChain(clusterParams[unique]);
        for (int i=0;i<unique;i++){
            for(int j=0;j<m;j++){
                if(cluster[j]==i){
                    clusterParams[i]=parameters[j];
                    clusterDists[i]=distributions[j];
                    break;
                }
            }
        }

        // Determine the weights for each cluster
        double[] clustWeights = new double[unique+1];
        clustWeights[unique] = mass;
        for(int i=0;i<unique;i++){
            clustWeights[i]=(double)clusterSize[i];
        }

        // Update clustering
        for (int i = 0; i < m; i++) {
            Angles[] currentData = data[i];
            // Remove current obs from its current cluster, and find cluster probs
            clustWeights[cluster[i]]=clustWeights[cluster[i]]-1;
            double[] tempProbs = new double[unique+1];
            for (int j = 0; j < unique+1; j++) {
                tempProbs[j] = clustWeights[j]*Math.exp(clusterDists[j].logDensity(currentData));
        //        System.out.print(tempProbs[j]+" ");
            }
        //    System.out.println();
            int index = randomInt(tempProbs,random);
   //         System.out.println(index);
            // Add observation to its new cluster.
            clustWeights[index]=clustWeights[index]+1;
            parameters[i]=clusterParams[index];
            distributions[i]=clusterDists[index];
            cluster[i]=index;
            if(index==unique){
                unique++;
                double[] tempClustWeights = new double[unique+1];
                ParameterSet[][] tempClustParams = new ParameterSet[unique+1][];
                VonMisesChain[] tempClustDists = new VonMisesChain[unique+1];
                for(int j=0;j<unique;j++){
                    tempClustParams[j]=clusterParams[j];
                    tempClustWeights[j]=clustWeights[j];
                    tempClustDists[j]=clusterDists[j];
                }
                clusterParams=tempClustParams;
                clustWeights=tempClustWeights;
                clusterDists=tempClustDists;
                clustWeights[unique-1]=1;
                clustWeights[unique]=mass;
            }
            clusterParams[unique]=prior.sample(random);
            clusterDists[unique]= new VonMisesChain(clusterParams[unique]);

        }
        // Updates cluster array and unique cluster count.
        countUnique();
        clusterIteration++;
    }



    /**
     * Draw sample parameters from the current state of the Markov chain
     */
    public ParameterSet[] sample(){
        double[] temp = new double[parameters.length+1];
        for(int i=0;i<parameters.length;i++){
            temp[i]=1.0;
        }
        temp[parameters.length]=mass;
        int index = randomInt(temp,random);
        if(index==parameters.length){
            return(prior.sample(random));
        }
        else{

            return(parameters[index]);
        }
    }

  /**
   * Determines number of unique clusters and updates unique variable accordingly. 
   * Also updates cluster array, and cluster size array
   */
     private void countUnique() {

         // Running count of unique clusters. A unique obsevation is the first 
        // time a given cluster appears.
        int temp = 0;
        // For each data point
        for (int i = 0; i < parameters.length; i++) {
            boolean priorIsSame = false;
            // Check if any preceding data point is identical (pointer identical)
            for (int j = 0; j < i; j++) {
                if (parameters[i] == parameters[j]) {
                    cluster[i] = cluster[j];
                    priorIsSame = true;
                    break;
                }
            }
            if (!priorIsSame) {
                cluster[i] = temp;
                temp++;
            }
        }

        unique = temp;
        
                
        clusterSize = new int[unique];
        for (int i = 0; i < unique; i++) {
            clusterSize[i] = 0;
        }
        for (int info : cluster) {
            clusterSize[info]++;
        }
                
    //   for(int i=0;i<unique;i++){
    //       System.out.print(clusterSize[i]+" ");
    //   }

    //   System.out.println(unique);
    }

    
    /**
     * Method takes an array of probabilities and returns an integer with probability corresponding
     * to the value at that position in the array
     */
    private int randomInt(double[] probArray, Random random) {
        // Clone probability array. Allows us to ensure that entries sum to one. Find scaling factor, and check that all
        // values are > 0
        double[] tempProbs = (double[]) (probArray.clone());
        double total = 0.0;
        for (int i = 0; i < tempProbs.length; i++) {
            if (tempProbs[i] < 0) {
                throw new IllegalArgumentException("Elements of probability array must be non-negative.");
            }
            total += tempProbs[i];
        }
        // Rescale probabilities to sum to 1, and change from density cdf
        tempProbs[0] /= total;
        for (int i = 1; i < tempProbs.length; i++) {
            tempProbs[i] = tempProbs[i] / total + tempProbs[i - 1];
        }

        double value = random.nextDouble();
        int index = 0;
        while (tempProbs[index] <= value) {
            index++;
        }
        return (index);
    }

}
