/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP;

import HMMDP.BaseDistributions.Angles;
import HMMDP.BaseDistributions.BivariateWishart;
import HMMDP.BaseDistributions.DistributionArray;
import HMMDP.BaseDistributions.Scale;
import java.util.Random;

/**
 *  Gives a chain of n independent Wishart distributions
 */
public class IndepWishartChain implements DistributionArray<Scale>{

    private BivariateWishart[] componentDists;
    private final int length;
    
    /**
     *  Initialize chain as i.i.d. Wisharts
     */
    public IndepWishartChain(BivariateWishart base, int n){
        this.length=n;
        componentDists = new BivariateWishart[n]; 
        for(int i=0;i<n;i++){
            componentDists[i]=base;
        }
    }
    
    
    public IndepWishartChain(BivariateWishart[] base){
        this.length=base.length;
        componentDists = new BivariateWishart[length];
        for(int i=0;i<length;i++){
            componentDists[i]=base[i];
        }
    }
    
    public double logDensity(Scale[] x) throws Exception {
        return logDensityKernel(x)+logNormalizingConstant();
    }
    
    /**
     * Calculates the full conditional likelihood of a scale chain
     */ 
    public double logLikelihood(Scale[] x, Angles[][] data, Angles[] centers) throws Exception{
        VonMisesChain condDist=new VonMisesChain(centers,x);
        double likelihood = 0.0;
        for(int i=0;i<data.length;i++){
            likelihood+=condDist.logDensity(data[i]);
        }
        return logDensity(x)+likelihood;
    }
    
    
    public double logDensityKernel(Scale[] x) throws Exception {
        if(x.length!=length) throw new Exception("Invalid Scale[] length.");
        double temp=0.0;
        for(int i=0;i<x.length;i++){
            temp+=(componentDists[i]).logDensityKernel(x[i]);
        }
        return temp;
    }

    public double logNormalizingConstant() {
        double temp=0.0;
        for(int i=0;i<length;i++){
            temp+=(componentDists[i]).logNormalizingConstant();
        }
        return temp;
    }

    public Scale[] sample(Random random) {
        Scale[] temp = new Scale[length];
        for(int i=0;i<length;i++){
            temp[i]=componentDists[i].sample(random);
        }
        return temp;
    }
    
    public int getLength(){
        return length;
    }
    
    
    /**
     *  Replace a single position in the chain with a new Wishart distribution
     */
    public void replacePosition(int pos, BivariateWishart newDist){
        componentDists[pos]=newDist;
    }

    public BivariateWishart reportPosition(int pos){
        return componentDists[pos];
    }
}
