/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
package HMMDP;

import HMMDP.BaseDistributions.Angles;
import HMMDP.BaseDistributions.BivariateVonMises;
import HMMDP.BaseDistributions.BivariateWishart;
import HMMDP.BaseDistributions.Scale;
import java.io.*;
import java.util.Random;

/**
 *
 * @author lennox
 */
public class Main {

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) throws FileNotFoundException, IOException, Exception {
        // Input file: temporarily hardcoded, will eventually be in args.
    //   String inputName="/home/lennox/Research/Protein/torgles/trunk/cortorgles/demo.in";
       if(!(args.length==1||args.length==2)){
           System.out.println("Usage: java -jar HMMDP.jar input-file");
          // System.out.println("OR: java -jar HMMDP.jar alt-input-file alt-data-file");
           System.out.println("OR: DPMHMM input-file");
           System.exit(0);
       }
       File inputFile = new File(args[0]);
       BufferedReader input = new BufferedReader(new FileReader(inputFile));
       String line = input.readLine();
       // Cycle through comments to first actual input line.
       while((line.trim()).charAt(0)==('#')){
           line=input.readLine();
       }
       
       String dataLoc="";
       String tagFile="";
       
       // For input type 1: positions files
       if(args.length==1){
         // Read in data location.
         dataLoc = (line.split("#")[0]).trim();
         line=readNext(input);       
         // Read in tag file name.
         tagFile = (line.split("#")[0]).trim();
         line=readNext(input);       
       } else{
       // For input type2: data file
           dataLoc=args[1];
       }
       
       // Read in output file names. 
       File outputFile = new File((line.split("#")[0]).trim());
       line=readNext(input);       
       File outputFile2 = new File((line.split("#")[0]).trim());
       line=readNext(input);       
       // Read in Markov chain settings
       int burnin = Integer.parseInt((line.split("#")[0]).trim());
       line=readNext(input);       
       int total = Integer.parseInt((line.split("#")[0]).trim());
       line=readNext(input);       
       int thinning = Integer.parseInt((line.split("#")[0]).trim());
       line=readNext(input);       
       int seed1 = Integer.parseInt((line.split("#")[0]).trim());
       line=readNext(input);       

       // Determine if informative (!=0) or noninformative (0) prior will be used. Determine if amino acid specific prior is used.
       boolean infPrior=false;
       boolean aas=false;
       int[] aminoArray=new int[1];
       int priorType;
       try{
           priorType=Integer.parseInt(line.split("#")[0].trim());
       }
       catch(NumberFormatException e){
           priorType=-1;
       }
       if(priorType==0){
           System.out.println("#Noninformative prior.");
       }
       else if (priorType==1){
           infPrior=true;
           System.out.println("#Informative prior.");
       } 
       else if(priorType==-1){
           infPrior=true;
           aas =true;
           String aaString[] = (line.split("#")[0]).split(",");
           aminoArray = new int[aaString.length];
           for(int i=0;i<aminoArray.length;i++){
               if(aaString[i].trim().equals("G"))
                   aminoArray[i]=2;
               else if(aaString[i].trim().equals("P"))
                   aminoArray[i]=1;
               else
                   aminoArray[i]=0;
               //System.out.print(aminoArray[i]);
           }
       }
       else {
         System.out.println("Prior type must be 0 (noninf), 1 (inf), or a comma separated amino acid list.");
         System.exit(0);
       }
       

       System.out.println("# Random seed: "+seed1);
       // Read in list of positions to be modeled.
       String positionsString="";
       if(args.length==1){
         line=readNext(input);       
         positionsString=(line.split("#")[0]).trim();
       } 
        
       //Entering mass param for DPM prior
       line=readNext(input);
       double tau = Double.parseDouble((line.split("#")[0]).trim());
       
       //Entering params for Wishart prior
       // Shape param
       line=readNext(input);
       double alpha = Double.parseDouble((line.split("#")[0]).trim());
       // Scale param. Scale matrix is this number times I.
       line=readNext(input);
       double wishDouble = Double.parseDouble((line.split("#")[0]).trim());
      
       
        Angles[][] data;
        if(args.length==1)               
          data= readData(dataLoc, tagFile);
        else
          data=readDataAlt(dataLoc,true);  

        int m = data.length;
        int n = data[0].length;
       
       
        // For reduced dataset (only available w/ input type 1
        if(positionsString.length()!=0){
          String[] positions=positionsString.split(",");        
          Angles[][] babyData = new Angles[m][positions.length];
          for(int i=0;i<m;i++){
            for(int j=0;j<(positions.length);j++){
              babyData[i][j]=data[i][Integer.parseInt(positions[j].trim())-1];
            }
          }
     
          data=babyData;
          n=babyData[0].length;
        }
        // Shouldn't happen anymore. A blank line now means model all positions.
        if(n==0){
            System.out.println("#Error. No positions selected for modeling");
            System.exit(0);
        }
            

        if(aas&&(n!=aminoArray.length)){
            System.out.println("Amino acid list length does not match number of positions.");
            System.exit(0);
        }
        
        
         // Initialize model
        System.out.println("#Initializing model with " + m + " peptides containing " + n + " alignment positions.");
        System.out.println("#"+burnin+" burnin iterations. "+total+" post burnin iterations.");
        if(aas){
            System.out.print("#Position specific priors: ");
            for(int i=0;i<aminoArray.length;i++) System.out.print(aminoArray[i]+" ");
            System.out.println();
        }
        Random myRandom = new Random(seed1);
        // Initializing model
        double[] initProbs;
        double[][] transProbs;
        BivariateVonMises[] states;
        
        // Test setup, noninformative vonMises prior
        if(!infPrior){
          initProbs = new double[1];
          initProbs[0]=1.0;
          transProbs= new double [1][1];   
          transProbs[0][0] = 1.0;
          states = new BivariateVonMises[1];
          states[0] = new BivariateVonMises(new Angles(0,0),Scale.precision(.1,.1,0));
        }
        else {
          // Set up the inital probability matrix from PDB data
          // 0 = turn (T)
          // 1 = coil (C)
          // 2 = helix (H)
          // 3 = sheet (E)
          initProbs = new double[4];
          initProbs[0]=0.1169;
          initProbs[1]=0.2037;
          initProbs[2]=0.3549;
          initProbs[3]=0.3245;
        
          // Set up transition probability matrix: row = current, column = next
          transProbs = new double[4][];
          double[] temp = {0.52183, 0.22963, 0.04993, 0.19861};
          transProbs[0] = temp;
          double[] temp2 = {0.10105, 0.48621, 0.11495, 0.29779};
          transProbs[1]=temp2;
          double[] temp3 =  {0.05730, 0.02887, 0.89217, 0.02166};
          transProbs[2] = temp3;
          double[] temp4 =  {0.04618, 0.20813, 0.02781, 0.71788};
          transProbs[3]= temp4;
          states = HiddenMarkovModelPeptide.generalInit();       
        }
        HiddenMarkovModelMean hmm;
        if(!aas){
            hmm = new HiddenMarkovModelMean(n, states, initProbs, transProbs);
        } else {
            hmm = new HiddenMarkovModelPeptide(n,initProbs,transProbs,aminoArray);
        }
        
        // Priors for scale parameter
        // Lennox et al scale prior
        Scale wishScale = Scale.precision(wishDouble,wishDouble,0);
        // Alternate scale prior
 //       Scale wishScale = Scale.parse(".1 .1 0");
        IndepWishartChain scaleChain = new IndepWishartChain(new BivariateWishart(alpha, wishScale), n);

        HiddenMarkovModelFull prior = new HiddenMarkovModelFull((HiddenMarkovModelMean)hmm, scaleChain);


        // Initialize DP model
        HmmClustering dp = new HmmClustering(data, tau, prior, myRandom, true);
        HmmClustering dp2 = new HmmClustering(data, tau, prior, myRandom, false);
        System.out.println("#Burnin iterations.");
        // burnin
        for (int i = 1; i <= burnin; i++) {
            dp.update();
            dp2.update();
            if(i%thinning==0){
              dp.report();
              dp2.report();
            }
        }
        // Evaluations
       
       DataOutputStream out1 = new DataOutputStream(new FileOutputStream(outputFile));
       DataOutputStream out2 = new DataOutputStream(new FileOutputStream(outputFile2));
       // Print header info for turnip library:
       // #of structs in file (1) # of residues (varys) # of targets (0) # Held out struct # (0)
       writeHeader(out1,n);
       writeHeader(out2,n);
        System.out.println("#Beginning grid evaluations.");
       for (int i = 1; i <= total; i++) {
            dp.update();
            dp2.update();
            if(i%thinning==0){
              dp.report();
              writeData(out1,dp.sample());
              dp2.report();
              writeData(out2,dp2.sample());
            }
        }
       
       out1.close();
       out2.close();
       
       // makeData(prior,myRandom);
    }

    private static Angles[][] readData(String dataLoc, String tagFile) throws FileNotFoundException, IOException {
        System.out.println("#"+dataLoc +" "+ tagFile);
        File pdbFile = new File(tagFile);
        BufferedReader input = new BufferedReader(new FileReader(pdbFile));
        String line = input.readLine();

        // Count tags = # of peptides
        int m = 0;
        while (line != null) {
            m++;
            line = input.readLine();
        }
        input.close();

        // Count number of positions - assumes data already cleaned so that all
        // tag.positions files contain every alignment position
        input = new BufferedReader(new FileReader(pdbFile));
        line = input.readLine();
        File dataFile = new File(dataLoc + line + ".positions");
        BufferedReader dataInput = new BufferedReader(new FileReader(dataFile));
        int n = 0;
        // Throw away header line, and read first line
        dataInput.readLine();
        String dataLine = dataInput.readLine();
        while (dataLine != null) {
            n++;
            dataLine = dataInput.readLine();
        }
        dataInput.close();

        
        
        // Set up data array. Rows are peptides, columns are positions
        Angles[][] data = new Angles[m][n];

        // Initialize data array
        String[] parsedLine;
        int mCount = 0;
        int nCount = 0;
        while (line != null) {
            dataInput = new BufferedReader(new FileReader(new File(dataLoc + line + ".positions")));
            // Discard header and read in first line
            dataInput.readLine();
            dataLine = dataInput.readLine();
            while (dataLine != null) {
                parsedLine = dataLine.split("\t");
                //System.out.println(parsedLine[1]+" "+parsedLine[2]);
                // If no data at location, create empty Angles object
                if (parsedLine[1].equals("-")) {
                    data[mCount][nCount] = new Angles();
                } // Otherwise read in phi, psi to Angles object. Note that data is in degrees, and use 
                // appropriate constructor.
                else {
                    data[mCount][nCount] = new Angles(Double.parseDouble(parsedLine[1]), Double.parseDouble(parsedLine[2]), true);
                }

                dataLine = dataInput.readLine();
                nCount++;
            }

            line = input.readLine();
            nCount = 0;
            mCount++;
        }
        return data;

    }

    // Reads data from loop library input files, no delimiters, no missing data allowed
    private static Angles[][] readDataAlt(String dataFile, boolean inDegrees) throws FileNotFoundException, IOException {
        System.out.println("#"+dataFile);
        BufferedReader input = new BufferedReader(new FileReader(dataFile));
        String line = input.readLine();
        // Count # of peptides m
        // and   # of positions n
        int m = 0;
        // Stops at EOF of first blank line.
        while ((line != null)&&((line.trim()).length()!=0)) {
            m++;
            line = input.readLine();
        }
        input.close();

        input = new BufferedReader(new FileReader(dataFile));
        line = input.readLine();

        String[] parsedLine = (line.trim()).split(" ");
        // Set up data array. Rows are peptides, columns are positions
        Angles[][] data = new Angles[m][(parsedLine.length)/2];        
        // Initialize data array
        int mCount = 0;
        while (line != null&&line.trim().length()!=0) {
            parsedLine = (line.trim()).split(" ");
            for (int nCount=0;nCount<parsedLine.length/2;nCount++) {
                data[mCount][nCount] = new Angles(Double.parseDouble(parsedLine[2*nCount]), Double.parseDouble(parsedLine[2*nCount+1]),inDegrees);
            }
            mCount++;

            line = input.readLine();
        }
        input.close();

        return data;
    }
    
    private static Angles[][] readFakeData(String dataFile, boolean inDegrees) throws FileNotFoundException, IOException {
        BufferedReader input = new BufferedReader(new FileReader(dataFile));
        String line = input.readLine();
        // Count # of peptides m
        // and   # of positions n
        int m = 0;
        while (line != null) {
            m++;
            line = input.readLine();
        }
        input.close();

        input = new BufferedReader(new FileReader(dataFile));
        line = input.readLine();

        String[] parsedLine = line.split(",");
        // Set up data array. Rows are peptides, columns are positions
        Angles[][] data = new Angles[m][parsedLine.length];

        // Initialize data array
        int mCount = 0;
        int nCount = 0;
        while (line != null) {
            parsedLine = line.split(",");
            for (String point : parsedLine) {
                String[] pointArray = point.split(" ");
                data[mCount][nCount] = new Angles(Double.parseDouble(pointArray[0]), Double.parseDouble(pointArray[1]),inDegrees);
                nCount++;
            }
            nCount = 0;
            mCount++;

            line = input.readLine();
        }
        input.close();

        return data;
    }

    private static Angles[][] readFakeData(String dataFile,int max) throws FileNotFoundException, IOException {
        BufferedReader input = new BufferedReader(new FileReader(dataFile));
        String line = input.readLine();
        // Count # of peptides m
        // and   # of positions n
        int m = 0;
        while (line != null) {
            m++;
            line = input.readLine();
        }
        input.close();

        input = new BufferedReader(new FileReader(dataFile));
        line = input.readLine();

        String[] parsedLine = line.split(",");
        // Set up data array. Rows are peptides, columns are positions
        if(parsedLine.length<max)
            max=parsedLine.length;
        System.out.println("#"+m + " " + max);
        Angles[][] data = new Angles[m][max];

        // Initialize data array
        int mCount = 0;
        int nCount = 0;
        while (line != null) {
            parsedLine = line.split(",");
            for (int i=0;i<max;i++) {
                String[] pointArray = parsedLine[i].split(" ");
                data[mCount][nCount] = new Angles(Double.parseDouble(pointArray[0]), Double.parseDouble(pointArray[1]));
                nCount++;
            }
            nCount = 0;
            mCount++;

            line = input.readLine();
        }
        input.close();

        return data;
    }
    
    private static Angles[] sample(HiddenMarkovModelFull prior, Random random) throws Exception {
        VonMisesChain sampleChain = new VonMisesChain(prior.sample(random));
        return (sampleChain.sample(random));
    }

    /**
     * Returns n samples from a single cluster.
     * @param prior
     * @param random
     * @param n
     * @return
     * @throws java.lang.Exception
     */
    private static Angles[][] sample(HiddenMarkovModelFull prior, Random random, int n) throws Exception {
        Angles[][] data = new Angles[n][];
        VonMisesChain sampleChain = new VonMisesChain(prior.sample(random));
        for (int i = 0; i < n; i++) {
            data[i] = sampleChain.sample(random);
        }
        return data;
    }


/**
 * Generates bin files
 */
  private static void writeData(DataOutputStream out,ParameterSet[] position) throws IOException, Exception {
    for ( ParameterSet p : position ) {
      out.writeDouble(p.center.phi());
      out.writeDouble(p.center.psi());
      out.writeDouble(p.scale.phi);
      out.writeDouble(p.scale.psi);
      out.writeDouble(p.scale.lambda);
    }
  }
  
  private static void writeHeader(DataOutputStream out, int length) throws Exception{
      out.writeInt(1);
      out.writeInt(length);
      out.writeInt(0);
      out.writeInt(0);
  }
  
  
  private static void makeData(HiddenMarkovModelFull prior, Random myRandom) throws Exception{
  
        for (int i = 0; i < 5; i++) {
            Angles[][] anglesSample = sample(prior, myRandom, 20);
            //  System.out.print("{");
            for (int j = 0; j < anglesSample.length; j++) {
                for (int k = 0; k < anglesSample[0].length; k++) {
                    System.out.print(anglesSample[j][k].toString(true) + ",");
                }
                //    System.out.print("new Angles("+anglesSample[j].phi()+","+anglesSample[j].psi()+"),");  
                System.out.println();
            }
        //  System.out.println("},");
        }
  }
  
  // Method to read the next line from the input file.
  private static String readNext(BufferedReader infile) throws IOException{
        String line=infile.readLine();
        while((line.trim()).charAt(0)==('#')){
          line=infile.readLine();
       }
       return line;
  }
  
}