/*
    Dirichlet Process Mixture of Hidden Markov Models: Accompanying Software
    Copyright (C) <2009>  <Kristin P. Lennox and David B. Dahl>

    Contact: lennox@stat.tamu.edu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

package HMMDP;


import HMMDP.BaseDistributions.Angles;
import HMMDP.BaseDistributions.BivariateVonMises;
import HMMDP.BaseDistributions.BivariateVonMisesMix;
import HMMDP.BaseDistributions.Distribution;
import HMMDP.BaseDistributions.DistributionArray;
import HMMDP.BaseDistributions.Scale;
import java.util.Random;

/**
 *  Assumes component distributions are independent, designed to work with von Mise distributions
 *  and finite mixtures of von Mises distributions.
 */

public class VonMisesChain implements DistributionArray<Angles>{

    private Distribution<Angles>[] componentDists;
    private final int length;

    /**
     *  Constructor that accept actual chains of distributions
     */
    public VonMisesChain(BivariateVonMises[] components){
        componentDists=components;
        this.length=components.length;
    }

     public VonMisesChain(ParameterSet[] components) throws Exception{
        componentDists=new BivariateVonMises[components.length];
        length=components.length;
        for(int i=0;i<length;i++){
            componentDists[i]=new BivariateVonMises(components[i]);
        }
    }
    
    public VonMisesChain(BivariateVonMisesMix[] components){
        componentDists=components;
        this.length=components.length;
    }
    
    /**
     * Constructor that generates chain from Angles and Scale objects
     */
    public VonMisesChain(Angles[] centers, Scale[] precisions) throws Exception{
        // Input verification
        if(centers.length!=precisions.length){
            throw new Exception("Parameter array lengths must match.");
        }
        // Initialize chain
        componentDists = new BivariateVonMises[centers.length];
        for(int i=0;i<centers.length;i++){
            // Check for valid parameters
            if(centers[i].isEmpty) throw new Exception("Invalid parameter array. Empty Angles object.");
            // If input is valid, initialize distribution
            componentDists[i]=new BivariateVonMises(centers[i],precisions[i]);
        }
        this.length=componentDists.length;
    }
    
    public double logDensity(Angles[][] x) throws Exception {
        return(logDensityKernel(x)+x.length*logNormalizingConstant());
    }
    
    public double logDensity(Angles[] x) throws Exception {
        return(logDensityKernel(x)+logNormalizingConstant());
    }

    public double logDensityKernel(Angles[][] x) throws Exception {
        double temp = 0.0;
        for(int i=0;i<x.length;i++){
            temp+=logDensityKernel(x[i]);
        }
        return temp;
    }
    
    /**
     *  Unobserved points do not contribute to density.
     */
    public double logDensityKernel(Angles[] x) throws Exception {
        if(x.length!=length) throw new Exception("Incorrect dimension.");
        double temp=0.0;
        for(int i=0; i<componentDists.length;i++){
            // Integrate over all possible values for unobserved points
            if(!x[i].isEmpty) 
                temp+=componentDists[i].logDensityKernel(x[i]);
        }
        return temp;

    }

    public double logNormalizingConstant() {
        double temp=0.0;
        for(int i=0; i<componentDists.length;i++){
            temp+=componentDists[i].logNormalizingConstant();
        }
        return temp;
    }

    public Angles[] sample(Random random) {
        Angles[] temp=new Angles[componentDists.length];
        for(int i=0;i<temp.length;i++){
            temp[i]=(Angles)componentDists[i].sample(random);
        }
        return(temp);
    }

    public int getLength() {
        return(length);
    }
    
    /**
     *  Replace a single position in the chain with a new distribution
     */
    public void replacePosition(int pos, Distribution<Angles> newDist){
        componentDists[pos]=newDist;
    }

    public Distribution reportPosition(int pos){
        return componentDists[pos];
    }
}
